# generated by "generate_qnames"
from sdc11073.namespaces import default_ns_helper

AbstractAlertReport = default_ns_helper.MSG.tag('AbstractAlertReport')
AbstractComponentReport = default_ns_helper.MSG.tag('AbstractComponentReport')
AbstractContextReport = default_ns_helper.MSG.tag('AbstractContextReport')
AbstractGet = default_ns_helper.MSG.tag('AbstractGet')
AbstractGetResponse = default_ns_helper.MSG.tag('AbstractGetResponse')
AbstractMetricReport = default_ns_helper.MSG.tag('AbstractMetricReport')
AbstractOperationalStateReport = default_ns_helper.MSG.tag('AbstractOperationalStateReport')
AbstractReport = default_ns_helper.MSG.tag('AbstractReport')
AbstractReportPart = default_ns_helper.MSG.tag('AbstractReportPart')
AbstractSet = default_ns_helper.MSG.tag('AbstractSet')
AbstractSetResponse = default_ns_helper.MSG.tag('AbstractSetResponse')
Activate = default_ns_helper.MSG.tag('Activate')
ActivateResponse = default_ns_helper.MSG.tag('ActivateResponse')
AlertState = default_ns_helper.MSG.tag('AlertState')
ArgValue = default_ns_helper.MSG.tag('ArgValue')
Argument = default_ns_helper.MSG.tag('Argument')
By = default_ns_helper.MSG.tag('By')
ComponentState = default_ns_helper.MSG.tag('ComponentState')
ContainmentTree = default_ns_helper.MSG.tag('ContainmentTree')
ContextState = default_ns_helper.MSG.tag('ContextState')
DescriptionModificationReport = default_ns_helper.MSG.tag('DescriptionModificationReport')
Descriptor = default_ns_helper.MSG.tag('Descriptor')
DescriptorRevisions = default_ns_helper.MSG.tag('DescriptorRevisions')
EpisodicAlertReport = default_ns_helper.MSG.tag('EpisodicAlertReport')
EpisodicComponentReport = default_ns_helper.MSG.tag('EpisodicComponentReport')
EpisodicContextReport = default_ns_helper.MSG.tag('EpisodicContextReport')
EpisodicMetricReport = default_ns_helper.MSG.tag('EpisodicMetricReport')
EpisodicOperationalStateReport = default_ns_helper.MSG.tag('EpisodicOperationalStateReport')
ErrorCode = default_ns_helper.MSG.tag('ErrorCode')
ErrorInfo = default_ns_helper.MSG.tag('ErrorInfo')
Filter = default_ns_helper.MSG.tag('Filter')
GetContainmentTree = default_ns_helper.MSG.tag('GetContainmentTree')
GetContainmentTreeResponse = default_ns_helper.MSG.tag('GetContainmentTreeResponse')
GetContextStates = default_ns_helper.MSG.tag('GetContextStates')
GetContextStatesByFilter = default_ns_helper.MSG.tag('GetContextStatesByFilter')
GetContextStatesByFilterResponse = default_ns_helper.MSG.tag('GetContextStatesByFilterResponse')
GetContextStatesByIdentification = default_ns_helper.MSG.tag('GetContextStatesByIdentification')
GetContextStatesByIdentificationResponse = default_ns_helper.MSG.tag('GetContextStatesByIdentificationResponse')
GetContextStatesResponse = default_ns_helper.MSG.tag('GetContextStatesResponse')
GetDescriptor = default_ns_helper.MSG.tag('GetDescriptor')
GetDescriptorResponse = default_ns_helper.MSG.tag('GetDescriptorResponse')
GetDescriptorsFromArchive = default_ns_helper.MSG.tag('GetDescriptorsFromArchive')
GetDescriptorsFromArchiveResponse = default_ns_helper.MSG.tag('GetDescriptorsFromArchiveResponse')
GetLocalizedText = default_ns_helper.MSG.tag('GetLocalizedText')
GetLocalizedTextResponse = default_ns_helper.MSG.tag('GetLocalizedTextResponse')
GetMdDescription = default_ns_helper.MSG.tag('GetMdDescription')
GetMdDescriptionResponse = default_ns_helper.MSG.tag('GetMdDescriptionResponse')
GetMdState = default_ns_helper.MSG.tag('GetMdState')
GetMdStateResponse = default_ns_helper.MSG.tag('GetMdStateResponse')
GetMdib = default_ns_helper.MSG.tag('GetMdib')
GetMdibResponse = default_ns_helper.MSG.tag('GetMdibResponse')
GetStatesFromArchive = default_ns_helper.MSG.tag('GetStatesFromArchive')
GetStatesFromArchiveResponse = default_ns_helper.MSG.tag('GetStatesFromArchiveResponse')
GetSupportedLanguages = default_ns_helper.MSG.tag('GetSupportedLanguages')
GetSupportedLanguagesResponse = default_ns_helper.MSG.tag('GetSupportedLanguagesResponse')
Handle = default_ns_helper.MSG.tag('Handle')
HandleRef = default_ns_helper.MSG.tag('HandleRef')
Identification = default_ns_helper.MSG.tag('Identification')
InvocationError = default_ns_helper.MSG.tag('InvocationError')
InvocationErrorMessage = default_ns_helper.MSG.tag('InvocationErrorMessage')
InvocationInfo = default_ns_helper.MSG.tag('InvocationInfo')
InvocationSource = default_ns_helper.MSG.tag('InvocationSource')
InvocationState = default_ns_helper.MSG.tag('InvocationState')
Lang = default_ns_helper.MSG.tag('Lang')
MdDescription = default_ns_helper.MSG.tag('MdDescription')
MdState = default_ns_helper.MSG.tag('MdState')
Mdib = default_ns_helper.MSG.tag('Mdib')
MetricState = default_ns_helper.MSG.tag('MetricState')
NumberOfLines = default_ns_helper.MSG.tag('NumberOfLines')
ObservedValueStream = default_ns_helper.MSG.tag('ObservedValueStream')
OperationHandleRef = default_ns_helper.MSG.tag('OperationHandleRef')
OperationInvokedReport = default_ns_helper.MSG.tag('OperationInvokedReport')
OperationState = default_ns_helper.MSG.tag('OperationState')
PeriodicAlertReport = default_ns_helper.MSG.tag('PeriodicAlertReport')
PeriodicComponentReport = default_ns_helper.MSG.tag('PeriodicComponentReport')
PeriodicContextReport = default_ns_helper.MSG.tag('PeriodicContextReport')
PeriodicMetricReport = default_ns_helper.MSG.tag('PeriodicMetricReport')
PeriodicOperationalStateReport = default_ns_helper.MSG.tag('PeriodicOperationalStateReport')
ProposedAlertState = default_ns_helper.MSG.tag('ProposedAlertState')
ProposedComponentState = default_ns_helper.MSG.tag('ProposedComponentState')
ProposedContextState = default_ns_helper.MSG.tag('ProposedContextState')
ProposedMetricState = default_ns_helper.MSG.tag('ProposedMetricState')
Ref = default_ns_helper.MSG.tag('Ref')
ReportPart = default_ns_helper.MSG.tag('ReportPart')
RequestedNumericValue = default_ns_helper.MSG.tag('RequestedNumericValue')
RequestedStringValue = default_ns_helper.MSG.tag('RequestedStringValue')
Retrievability = default_ns_helper.MSG.tag('Retrievability')
RetrievabilityInfo = default_ns_helper.MSG.tag('RetrievabilityInfo')
SetAlertState = default_ns_helper.MSG.tag('SetAlertState')
SetAlertStateResponse = default_ns_helper.MSG.tag('SetAlertStateResponse')
SetComponentState = default_ns_helper.MSG.tag('SetComponentState')
SetComponentStateResponse = default_ns_helper.MSG.tag('SetComponentStateResponse')
SetContextState = default_ns_helper.MSG.tag('SetContextState')
SetContextStateResponse = default_ns_helper.MSG.tag('SetContextStateResponse')
SetMetricState = default_ns_helper.MSG.tag('SetMetricState')
SetMetricStateResponse = default_ns_helper.MSG.tag('SetMetricStateResponse')
SetString = default_ns_helper.MSG.tag('SetString')
SetStringResponse = default_ns_helper.MSG.tag('SetStringResponse')
SetValue = default_ns_helper.MSG.tag('SetValue')
SetValueResponse = default_ns_helper.MSG.tag('SetValueResponse')
SourceMds = default_ns_helper.MSG.tag('SourceMds')
State = default_ns_helper.MSG.tag('State')
StateRevisions = default_ns_helper.MSG.tag('StateRevisions')
SystemErrorReport = default_ns_helper.MSG.tag('SystemErrorReport')
Text = default_ns_helper.MSG.tag('Text')
TextWidth = default_ns_helper.MSG.tag('TextWidth')
TimeFrame = default_ns_helper.MSG.tag('TimeFrame')
TransactionId = default_ns_helper.MSG.tag('TransactionId')
Value = default_ns_helper.MSG.tag('Value')
Version = default_ns_helper.MSG.tag('Version')
VersionFrame = default_ns_helper.MSG.tag('VersionFrame')
WaveformStream = default_ns_helper.MSG.tag('WaveformStream')



from .dispatchkey import DispatchKey, RequestDispatcher
from .messageconverter import MessageConverterMiddleware
from .pathelementregistry import PathElementRegistry
from .request import RequestData

__all__ = ['PathElementRegistry', 'RequestData', 'RequestDispatcher', 'DispatchKey', 'MessageConverterMiddleware']



"""Compression module for http."""
import contextlib
import zlib
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import ClassVar

try:
    import lz4.frame
except ImportError:
    lz4 = None


class CompressionError(Exception):
    pass


class AbstractDataCompressor(ABC):
    algorithms = ()

    @staticmethod
    @abstractmethod
    def compress_payload(payload):
        pass

    @staticmethod
    @abstractmethod
    def decompress_payload(payload):
        pass


class CompressionHandler:
    """Compression handler.
    Should be used by servers and clients that are supposed to handle compression.
    """

    available_encodings: ClassVar[list[str]] = []  # initial default
    handlers: ClassVar[dict[str, type[AbstractDataCompressor]]] = {}

    @classmethod
    def register_handler(cls, handler: type[AbstractDataCompressor]):
        for alg in handler.algorithms:
            if alg.lower() in cls.available_encodings:
                raise ValueError(f'Algorithm {alg} already registered, class = {cls.__name__} ')
        for alg in handler.algorithms:
            cls.available_encodings.append(alg.lower())
            cls.handlers[alg] = handler

    @classmethod
    def compress_payload(cls, algorithm: str, payload: bytes):
        """Compresses payload based on required algorithm.
        Raises CompressionException if algorithm is not supported.

        :param algorithm: one of strings provided by registered compression handlers
        :param payload: text to compress
        @return: compressed content
        """
        return cls.get_handler(algorithm).compress_payload(payload)

    @classmethod
    def decompress_payload(cls, algorithm: str, payload: bytes):
        """Decompresses payload based on required algorithm.
        Raises CompressionException if algorithm is not supported.

        :param algorithm: one of strings provided by registered compression handlers
        :param payload: text to decompress
        @return: decompressed content
        """
        return cls.get_handler(algorithm).decompress_payload(payload)

    @classmethod
    def get_handler(cls, algorithm: str):
        """:param algorithm: one of strings provided by registered compression handlers
        :return: AbstractDataCompressor implementation
        """
        handler = cls.handlers.get(algorithm.lower())
        if not handler:
            txt = f"{algorithm} compression is not supported. Only {cls.available_encodings} are supported."
            raise CompressionError(txt)
        return handler

    @staticmethod
    def parse_header(header):
        """Examples of headers are:  Examples of its use are:

        Accept-Encoding: compress, gzip
        Accept-Encoding:
        Accept-Encoding: *
        Accept-Encoding: compress;q=0.5, gzip;q=1.0
        Accept-Encoding: gzip;q=1.0, identity; q=0.5, *;q=0

        returns sorted list of compression algorithms by priority
        """
        # for now work with standard python containers
        # if performance becomes an issue could be done within one loop
        parsed_headers = OrderedDict()
        if header:
            for alg in (x.split(";") for x in header.split(",")):
                alg_name = alg[0].strip()
                parsed_headers[alg_name] = 1  # default
                with contextlib.suppress(ValueError, IndexError):
                    parsed_headers[alg_name] = float(alg[1].split("=")[1])

        return [pair[0] for pair in sorted(parsed_headers.items(), key=lambda kv: kv[1], reverse=True)]


class GzipCompressionHandler(AbstractDataCompressor):
    algorithms = ('gzip',)

    @staticmethod
    def compress_payload(payload: bytes):
        if not isinstance(payload, bytes):
            raise TypeError(f'a bytes-like object is required, not "{payload.__class__.__name__}", payload={payload}')
        gzip_compress = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, 16 + zlib.MAX_WBITS)
        return gzip_compress.compress(payload) + gzip_compress.flush()

    @staticmethod
    def decompress_payload(payload: bytes):
        return zlib.decompress(payload, 16 + zlib.MAX_WBITS)


CompressionHandler.register_handler(GzipCompressionHandler)


class Lz4CompressionHandler(AbstractDataCompressor):
    algorithms = ('x-lz4', 'lz4')

    @staticmethod
    def compress_payload(payload: bytes):
        return lz4.frame.compress(payload)

    @staticmethod
    def decompress_payload(payload: bytes):
        return lz4.frame.decompress(payload)


if lz4 is not None:
    CompressionHandler.register_handler(Lz4CompressionHandler)



"""The module contains the implementation of Consumermdib."""
from __future__ import annotations

import enum
import threading
import time
import traceback
from collections import deque
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable

from sdc11073 import loghelper
from sdc11073 import observableproperties as properties
from sdc11073.exceptions import ApiUsageError

from . import mdibbase
from .consumermdibxtra import ConsumerMdibMethods

if TYPE_CHECKING:
    from decimal import Decimal
    from enum import Enum

    from sdc11073.consumer import SdcConsumer
    from sdc11073.pysoap.msgreader import MdibVersionGroupReader
    from sdc11073.xml_types.msg_types import (
        AbstractReport,
        DescriptionModificationReport,
        EpisodicAlertReport,
        EpisodicComponentReport,
        EpisodicContextReport,
        EpisodicMetricReport,
        OperationInvokedReport,
    )

    from .entityprotocol import EntityGetterProtocol
    from .mdibbase import MdibVersionGroup
    from .statecontainers import (
        AbstractContextStateContainer,
        AbstractStateContainer,
        RealTimeSampleArrayMetricStateContainer,
    )


@dataclass
class RtSampleContainer:
    """RtSampleContainer contains a single value."""

    value: Decimal
    determination_time: float
    validity: Enum
    annotations: list

    @property
    def age(self) -> float:
        """Return the age of the Sample in seconds."""
        return time.time() - self.determination_time

    def __repr__(self) -> str:
        return f'RtSample value="{self.value}" validity="{self.validity}" time={self.determination_time}'


class ConsumerRtBuffer:
    """Collects data of one real time stream."""

    def __init__(self,
                 sample_period: float,
                 max_samples: int):
        """Construct a ConsumerRtBuffer.

        :param sample_period: float value, in seconds.
                              When an incoming real time sample array is split into single RtSampleContainers,
                              this is used to calculate the individual time stamps.
                              Value can be zero if correct value is not known.
                              In this case all Containers will have the observation time of the sample array.
        :param max_samples: integer, max. length of self.rt_data
        """
        self.rt_data = deque(maxlen=max_samples)
        self.sample_period = sample_period
        self._max_samples = max_samples
        self._logger = loghelper.get_logger_adapter('sdc.client.mdib.rt')
        self._lock = Lock()
        self.last_sc = None  # last state container that was handled

    def mk_rt_sample_containers(self, realtime_sample_array_container: RealTimeSampleArrayMetricStateContainer) \
            -> list[RtSampleContainer]:
        """Create a list of RtSampleContainer from a RealTimeSampleArrayMetricStateContainer.

        :param realtime_sample_array_container: a RealTimeSampleArrayMetricStateContainer instance
        :return: a list of RtSampleContainer
        """
        self.last_sc = realtime_sample_array_container
        metric_value = realtime_sample_array_container.MetricValue
        if metric_value is None:
            # this can happen if metric state is not activated.
            self._logger.debug('real time sample array "{} "has no metric value, ignoring it',  # noqa: PLE1205
                               realtime_sample_array_container.DescriptorHandle)
            return []
        determination_time = metric_value.DeterminationTime
        annots = metric_value.Annotation
        apply_annotations = metric_value.ApplyAnnotation
        rt_sample_containers = []
        if metric_value.Samples is not None:
            for i, sample in enumerate(metric_value.Samples):
                applied_annotations = []
                if apply_annotations is not None:
                    for apply_annotation in apply_annotations:
                        if apply_annotation.SampleIndex == i:
                            # there is an annotation for this sample:
                            ann_index = apply_annotation.AnnotationIndex
                            annotation = annots[ann_index]  # index is zero-based
                            applied_annotations.append(annotation)
                rt_sample_time = determination_time + i * self.sample_period
                rt_sample_containers.append(RtSampleContainer(sample,
                                                              rt_sample_time,
                                                              metric_value.MetricQuality.Validity,
                                                              applied_annotations))
        return rt_sample_containers

    def add_rt_sample_containers(self, rt_sample_containers: list[RtSampleContainer]) -> None:
        """Update self.rt_data with the new rt_sample_containers.

        :param rt_sample_containers: a list of RtSampleContainer
        :return: None
        """
        if not rt_sample_containers:
            return
        with self._lock:
            self.rt_data.extend(rt_sample_containers)

    def read_rt_data(self) -> list[RtSampleContainer]:
        """Consume all currently buffered data and return it.

        :return: a list of RtSampleContainer objects
        """
        with self._lock:
            ret = list(self.rt_data)
            self.rt_data.clear()
        return ret


@dataclass
class _BufferedData:
    mdib_version_group: MdibVersionGroupReader
    data: AbstractReport
    handler: callable


class ConsumerMdibState(enum.Enum):
    """ConsumerMdib can be in one of these states."""

    initializing = enum.auto()  # the state during reload_all()
    initialized = enum.auto()  # the state when mdib is in sync with provider
    invalid = enum.auto()  # the state when mdib is not in sync with provider


class ConsumerMdib(mdibbase.MdibBase):
    """ConsumerMdib is a mirror of a provider mdib. Updates are performed by an SdcConsumer."""

    # for testing purpose you can disable checking of mdib version, so that every notification is accepted.
    MDIB_VERSION_CHECK_DISABLED = False

    # sequence_or_instance_id_changed_event is set to True every time the sequence id changes.
    # It is not reset to False any time later.
    # It is in the responsibility of the application to react on a changed sequence id.
    # Observe this property and call "reload_all" in the observer code.
    sequence_or_instance_id_changed_event: bool = properties.ObservableProperty(
        default_value=False, fire_only_on_changed_value=False)

    def __init__(self,
                 sdc_client: SdcConsumer,
                 extras_cls: type | None = None,
                 max_realtime_samples: int = 100):
        """Construct a ConsumerMdib instance.

        :param sdc_client: a SdcConsumer instance
        :param  extras_cls: extended functionality
        :param max_realtime_samples: determines how many real time samples are stored per RealtimeSampleArray
        """
        super().__init__(sdc_client.sdc_definitions,
                         loghelper.get_logger_adapter('sdc.client.mdib', sdc_client.log_prefix))
        self._sdc_client = sdc_client
        if extras_cls is None:
            extras_cls = ConsumerMdibMethods
        self._xtra = extras_cls(self, self._logger)
        self._state = ConsumerMdibState.invalid
        self.rt_buffers = {}  # key  is a handle, value is a ConsumerRtBuffer
        self._max_realtime_samples = max_realtime_samples
        self._last_wf_age_log = time.time()
        # a buffer for notifications that are received before initial get_mdib is done
        self._buffered_notifications = []
        self._buffered_notifications_lock = Lock()
        self.entities: EntityGetterProtocol = mdibbase.EntityGetter(self)

    @property
    def xtra(self) -> Any:
        """Give access to extended functionality."""
        return self._xtra

    @property
    def sdc_client(self) -> SdcConsumer:
        """Give access to sdc client."""
        return self._sdc_client

    @property
    def is_initialized(self) -> bool:
        """Returns True if everything has been set up completely."""
        return self._state == ConsumerMdibState.initialized

    def init_mdib(self):
        """Binds own notification handlers to observables of sdc client and calls GetMdib.

        Client mdib is initialized from GetMdibResponse, and from then on updated from incoming notifications.
        :return: None
        """
        if self.is_initialized:
            raise ApiUsageError('ConsumerMdib is already initialized')
        # first start receiving notifications, then call get_mdib.
        # Otherwise, we might miss notifications.
        self._xtra.bind_to_client_observables()
        self.reload_all()
        self._sdc_client.set_mdib(self)  # pylint: disable=protected-access
        self._logger.info('initializing mdib done')

    def reload_all(self):
        """Delete all data and reloads everything."""
        self._logger.info('reload_all called')
        with self.mdib_lock:
            self._state = ConsumerMdibState.initializing  # notifications are now buffered
            self.descriptions.clear()
            self.clear_states()
            self.sequence_id = None
            self.instance_id = None
            self.mdib_version = None

            get_service = self._sdc_client.client('Get')
            self._logger.info('initializing mdib...')
            response = get_service.get_mdib()  # GetRequestResult
            self._logger.info('creating description containers...')
            descriptor_containers, state_containers = response.result
            self.add_description_containers(descriptor_containers)
            self._logger.info('creating state containers...')
            self.add_state_containers(state_containers)

            mdib_version_group = response.mdib_version_group
            self.mdib_version = mdib_version_group.mdib_version
            self._logger.info('setting initial mdib version to %d',
                              mdib_version_group.mdib_version)
            self.sequence_id = mdib_version_group.sequence_id
            self._logger.info('setting initial sequence id to {}', mdib_version_group.sequence_id)  # noqa: PLE1205
            if mdib_version_group.instance_id != self.instance_id:
                self.instance_id = mdib_version_group.instance_id
            self._logger.info('setting initial instance id to {}', mdib_version_group.instance_id)  # noqa: PLE1205

            # retrieve context states only if there were none in mdib
            if len(self.context_states.objects) == 0:
                self._get_context_states()
            else:
                self._logger.info('found context states in GetMdib Result, will not call getContextStates')

            # process buffered notifications
            with self._buffered_notifications_lock:
                self._logger.debug('got _buffered_notifications_lock')
                for buffered_report in self._buffered_notifications:
                    # buffered data might contain notifications that do not fit.
                    if buffered_report.mdib_version_group.sequence_id != self.sequence_id:
                        self.logger.debug('wrong sequence id "%s"; ignore buffered report',
                                          buffered_report.mdib_version_group.sequence_id)
                        continue
                    if buffered_report.mdib_version_group.mdib_version <= self.mdib_version:
                        self.logger.debug('older mdib version "%d"; ignore buffered report',
                                          buffered_report.mdib_version_group.mdib_version)
                        continue
                    buffered_report.handler(buffered_report.mdib_version_group,
                                            buffered_report.data)
                del self._buffered_notifications[:]
                self._state = ConsumerMdibState.initialized
            self._logger.info('reload_all done')

    def _get_context_states(self, handles: list[str] | None = None):
        try:
            self._logger.debug('new Query, handles={}', handles)  # noqa: PLE1205
            time.sleep(0.001)
            context_service = self._sdc_client.client('Context')
            self._logger.info('requesting context states...')
            response = context_service.get_context_states(handles)
            context_state_containers = response.result.ContextState

            self._logger.debug('got {} context states', len(context_state_containers))  # noqa: PLE1205
            with self.context_states.lock:
                for state_container in context_state_containers:
                    old_state_containers = self.context_states.handle.get(state_container.Handle, [])
                    if len(old_state_containers) == 0:
                        self.context_states.add_object_no_lock(state_container)
                        self._logger.debug('new ContextState {}', state_container)  # noqa: PLE1205
                    elif len(old_state_containers) == 1:
                        old_state_container = old_state_containers[0]
                        if old_state_container.StateVersion != state_container.StateVersion:
                            self._logger.debug('update {} ==> {}',  # noqa: PLE1205
                                               old_state_container, state_container)
                            old_state_container.update_from_node(state_container.node)
                            self.context_states.update_object_no_lock(old_state_container)
                        else:
                            difference = state_container.diff(old_state_container)
                            if difference:
                                self._logger.error('no state version update but different!\n{ \n{}',  # noqa: PLE1205
                                                   difference)
                    else:
                        txt = ', '.join([str(x) for x in old_state_containers])
                        self._logger.error('found {} objects: {}', len(old_state_containers), txt)  # noqa: PLE1205

        except Exception:  # noqa: BLE001
            self._logger.error(traceback.format_exc()) # noqa: TRY400
        finally:
            self._logger.info('_get_context_states done')

    def _can_accept_mdib_version(self, new_mdib_version: int, log_prefix: str) -> bool:
        if self.MDIB_VERSION_CHECK_DISABLED:
            return True
        # log deviations from expected mdib version
        if new_mdib_version < self.mdib_version:
            self._logger.warning('{}: ignoring too old Mdib version, have {}, got {}',  # noqa: PLE1205
                                 log_prefix, self.mdib_version, new_mdib_version)
        elif (new_mdib_version - self.mdib_version) > 1:
            # This can happen if consumer did not subscribe to all notifications.
            # Still log a warning, because mdib is no longer a correct mirror of provider mdib.
            self._logger.warning('{}: expect mdib_version {}, got {}',  # noqa: PLE1205
                                 log_prefix, self.mdib_version + 1, new_mdib_version)
        # it is possible to receive multiple notifications with the same mdib version => compare ">="
        return new_mdib_version >= self.mdib_version

    def _check_sequence_or_instance_id_changed(self, mdib_version_group: MdibVersionGroup):
        """Check if sequence id and instance id are still the same.

        If not,
        - set state member to invalid
        - set the observable "sequence_or_instance_id_changed_event" in a thread.
          This allows to implement an observer that can directly call reload_all without blocking the consumer.
        """
        if mdib_version_group.sequence_id == self.sequence_id and mdib_version_group.instance_id == self.instance_id:
            return
        if self._state == ConsumerMdibState.initialized:
            if mdib_version_group.sequence_id != self.sequence_id:
                self.logger.warning('sequence id changed from "%s" to "%s"',
                                    self.sequence_id, mdib_version_group.sequence_id)
            if mdib_version_group.instance_id != self.instance_id:
                self.logger.warning('instance id changed from "%r" to "%r"',
                                    self.instance_id, mdib_version_group.instance_id)
            self.logger.warning('mdib is no longer valid!')

            self._state = ConsumerMdibState.invalid

            def _set_observable():
                self.sequence_or_instance_id_changed_event = True

            thr = threading.Thread(target=_set_observable)
            thr.start()

    def _update_from_mdib_version_group(self, mdib_version_group: MdibVersionGroupReader):
        if mdib_version_group.mdib_version != self.mdib_version:
            self.mdib_version = mdib_version_group.mdib_version
        if mdib_version_group.sequence_id != self.sequence_id:
            self.sequence_id = mdib_version_group.sequence_id
        if mdib_version_group.instance_id != self.instance_id:
            self.instance_id = mdib_version_group.instance_id

    def _update_from_states_report(self, report_type: str,
                                   report: (EpisodicMetricReport | EpisodicAlertReport | OperationInvokedReport
                                            | EpisodicComponentReport),
                                   ) -> dict[str, AbstractStateContainer]:
        """Update mdib with incoming states."""
        states_by_handle = {}
        for report_part in report.ReportPart:
            for state_container in report_part.values_list:
                src = self.states
                old_state_container = src.descriptor_handle.get_one(state_container.DescriptorHandle,
                                                                    allow_none=True)
                if old_state_container is not None:
                    if self._has_new_state_usable_state_version(old_state_container, state_container,
                                                                report_type):
                        old_state_container.update_from_other_container(state_container)
                        src.update_object(old_state_container)
                        states_by_handle[old_state_container.DescriptorHandle] = old_state_container
                else:
                    self._logger.error('{}: got a new state {}',  # noqa: PLE1205
                                       report_type,
                                       state_container.DescriptorHandle)
                    self._set_descriptor_container_reference(state_container)
                    src.add_object(state_container)
                    states_by_handle[state_container.DescriptorHandle] = state_container
        return states_by_handle

    def _update_from_context_states_report(self, report: EpisodicContextReport,
                                           ) -> dict[str, AbstractContextStateContainer]:
        """Update mdib with incoming states."""
        states_by_handle = {}
        for report_part in report.ReportPart:
            for state_container in report_part.values_list:
                src = self.context_states
                old_state_container = src.handle.get_one(state_container.Handle,
                                                         allow_none=True)
                if old_state_container is not None:
                    if self._has_new_state_usable_state_version(old_state_container, state_container,
                                                                'context states'):
                        self._logger.info(  # noqa: PLE1205
                            'updated context state: handle = {} Descriptor Handle={} Assoc={}, Validators={}',
                            state_container.Handle, state_container.DescriptorHandle,
                            state_container.ContextAssociation, state_container.Validator)
                        old_state_container.update_from_other_container(state_container)
                        src.update_object(old_state_container)
                        states_by_handle[old_state_container.DescriptorHandle] = old_state_container
                else:
                    self._logger.info(  # noqa: PLE1205
                        'new context state: handle = {} Descriptor Handle={} Assoc={}, Validators={}',
                        state_container.Handle, state_container.DescriptorHandle,
                        state_container.ContextAssociation, state_container.Validator)
                    self._set_descriptor_container_reference(state_container)
                    src.add_object(state_container)
                    states_by_handle[state_container.Handle] = state_container
        return states_by_handle

    def _pre_check_report_ok(self,
                             mdib_version_group: MdibVersionGroupReader,
                             report: AbstractReport | list[RealTimeSampleArrayMetricStateContainer],
                             handler: Callable) -> bool:
        """Check if the report can be added to mdib.

        The pre-check runs before the mdib lock is acquired.
        The report is buffered if state is 'initializing' and 'is_buffered_report' is False.
        :return: True if report can be added to mdib.
        """
        self._check_sequence_or_instance_id_changed(mdib_version_group)  # this might change self._state
        if self._state == ConsumerMdibState.invalid:
            # ignore report in these states
            return False
        if self._state == ConsumerMdibState.initializing:
            with self._buffered_notifications_lock:
                # check state again, it might have changed before lock was acquired
                if self._state == ConsumerMdibState.initializing:
                    self._buffered_notifications.append(_BufferedData(mdib_version_group, report, handler))
                    return False
        return True

    def process_incoming_metric_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                              report: EpisodicMetricReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self._process_incoming_metric_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_metric_states_report(mdib_version_group, report)

    def _process_incoming_metric_states_report(self,
                                               mdib_version_group: MdibVersionGroupReader,
                                               report: EpisodicMetricReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'metric states'):
                self._update_from_mdib_version_group(mdib_version_group)
                states_by_handle = self._update_from_states_report('metric states', report)
        finally:
            self.metrics_by_handle = states_by_handle  # update observable

    def process_incoming_alert_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                             report: EpisodicAlertReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self._process_incoming_alert_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_alert_states_report(mdib_version_group, report)

    def _process_incoming_alert_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                              report: EpisodicAlertReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'alert states'):
                self._update_from_mdib_version_group(mdib_version_group)
                states_by_handle = self._update_from_states_report('alert states', report)
        finally:
            self.alert_by_handle = states_by_handle  # update observable

    def process_incoming_operational_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                                   report: OperationInvokedReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self.process_incoming_operational_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_operational_states_report(mdib_version_group, report)

    def _process_incoming_operational_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                                    report: OperationInvokedReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'operational states'):
                self._update_from_mdib_version_group(mdib_version_group)
                states_by_handle = self._update_from_states_report('operational states', report)
        finally:
            self.operation_by_handle = states_by_handle  # update observable

    def process_incoming_context_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                               report: EpisodicContextReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self._process_incoming_context_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_context_states_report(mdib_version_group, report)

    def _process_incoming_context_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                                report: EpisodicContextReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'context states'):
                self._update_from_mdib_version_group(mdib_version_group)
                states_by_handle = self._update_from_context_states_report(report)
        finally:
            self.context_by_handle = states_by_handle  # update observable

    def process_incoming_component_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                                 report: EpisodicComponentReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self._process_incoming_component_states_report):
            return
        with self.mdib_lock:
            self._process_incoming_component_states_report(mdib_version_group, report)

    def _process_incoming_component_states_report(self, mdib_version_group: MdibVersionGroupReader,
                                                  report: EpisodicComponentReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'component states'):
                self._update_from_mdib_version_group(mdib_version_group)
                states_by_handle = self._update_from_states_report('component states', report)
        finally:
            self.component_by_handle = states_by_handle  # update observable

    def process_incoming_waveform_states(self, mdib_version_group: MdibVersionGroupReader,
                                         state_containers: list[RealTimeSampleArrayMetricStateContainer],
                                         ) -> dict[str, RealTimeSampleArrayMetricStateContainer] | None:
        """Check mdib_version_group and process state_containers it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, state_containers,
                                         self._process_incoming_waveform_states):
            return
        with self.mdib_lock:
            self._process_incoming_waveform_states(mdib_version_group, state_containers)

    def _process_incoming_waveform_states(self, mdib_version_group: MdibVersionGroupReader,
                                          state_containers: list[RealTimeSampleArrayMetricStateContainer]):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observable.
        Call this method only if mdib_lock is already acquired.
        """
        states_by_handle = {}
        try:
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'waveform states'):
                self._update_from_mdib_version_group(mdib_version_group)
                for state_container in state_containers:
                    old_state_container = self.states.descriptor_handle.get_one(state_container.DescriptorHandle,
                                                                                allow_none=True)
                    if old_state_container is not None:
                        if self._has_new_state_usable_state_version(old_state_container, state_container,
                                                                    'waveform states'):
                            old_state_container.update_from_other_container(state_container)
                            self.states.update_object(old_state_container)
                            states_by_handle[old_state_container.DescriptorHandle] = old_state_container
                    else:
                        self._logger.error('waveform states: got a new state {}',  # noqa: PLE1205
                                           state_container.DescriptorHandle)
                        self._set_descriptor_container_reference(state_container)
                        self.states.add_object(state_container)
                        states_by_handle[state_container.DescriptorHandle] = state_container

                # add to Waveform Buffer
                for state_container in states_by_handle.values():
                    state_container: RealTimeSampleArrayMetricStateContainer
                    descriptor_container = state_container.descriptor_container
                    d_handle = state_container.DescriptorHandle
                    rt_buffer = self.rt_buffers.get(d_handle)
                    if rt_buffer is None:
                        sample_period = 0  # default
                        if descriptor_container is not None:
                            # read sample period
                            sample_period = descriptor_container.SamplePeriod or 0
                        rt_buffer = ConsumerRtBuffer(sample_period=sample_period,
                                                     max_samples=self._max_realtime_samples)
                        self.rt_buffers[d_handle] = rt_buffer
                    rt_sample_containers = rt_buffer.mk_rt_sample_containers(state_container)
                    rt_buffer.add_rt_sample_containers(rt_sample_containers)
        finally:
            self.waveform_by_handle = states_by_handle  # update observable

    def process_incoming_description_modifications(self, mdib_version_group: MdibVersionGroupReader,
                                                   report: DescriptionModificationReport):
        """Check mdib_version_group and process report it if okay."""
        if not self._pre_check_report_ok(mdib_version_group, report,
                                         self._process_incoming_description_modifications):
            return
        with self.mdib_lock:
            self._process_incoming_description_modifications(mdib_version_group, report)

    def _process_incoming_description_modifications(self, #  noqa: PLR0915, PLR0912, C901
                                                    mdib_version_group: MdibVersionGroupReader,
                                                    report: DescriptionModificationReport):
        """Check mdib version.

        If okay:
         - update mdib.
         - update observables.
        Call this method only if mdib_lock is already acquired.
        """

        def multi_key(st_container: AbstractStateContainer) -> mdibbase.StatesLookup | mdibbase.MultiStatesLookup:
            return self.context_states if st_container.is_context_state else self.states

        new_descriptor_by_handle = {}
        updated_descriptor_by_handle = {}
        deleted_descriptor_by_handle = {}
        try:
            dmt = self.sdc_definitions.data_model.msg_types.DescriptionModificationType
            if self._can_accept_mdib_version(mdib_version_group.mdib_version, 'descriptors'):
                self._update_from_mdib_version_group(mdib_version_group)
                for report_part in report.ReportPart:
                    modification_type = report_part.ModificationType
                    if modification_type == dmt.CREATE:
                        for descriptor_container in report_part.Descriptor:
                            self.descriptions.add_object(descriptor_container)
                            self._logger.debug(  # noqa: PLE1205
                                'process_incoming_descriptors: created description "{}" (parent="{}")',
                                descriptor_container.Handle, descriptor_container.parent_handle)
                            new_descriptor_by_handle[descriptor_container.Handle] = descriptor_container
                        for state_container in report_part.State:
                            self._set_descriptor_container_reference(state_container)
                            multi_key(state_container).add_object_no_lock(state_container)
                    elif modification_type == dmt.UPDATE:
                        updated_descriptor_containers = report_part.Descriptor
                        updated_state_containers = report_part.State
                        for descriptor_container in updated_descriptor_containers:
                            self._logger.info(  # noqa: PLE1205
                                'process_incoming_descriptors: update descriptor "{}" (parent="{}")',
                                descriptor_container.Handle, descriptor_container.parent_handle)
                            old_container = self.descriptions.handle.get_one(descriptor_container.Handle,
                                                                             allow_none=True)
                            if old_container is None:
                                self._logger.error(  # noqa: PLE1205
                                    'process_incoming_descriptors: got update of descriptor "{}", '
                                    'but it did not exist in mdib!',
                                    descriptor_container.Handle)
                            else:
                                old_container.update_from_other_container(descriptor_container)
                            updated_descriptor_by_handle[descriptor_container.Handle] = descriptor_container
                            # if this is a context descriptor, delete all associated states that are not in
                            # state_containers list
                            if descriptor_container.is_context_descriptor:
                                updated_handles = {s.Handle for s in updated_state_containers
                                                   if s.DescriptorHandle == descriptor_container.Handle}
                                my_handles = {s.Handle for s in self.context_states.descriptor_handle.get(
                                    descriptor_container.Handle, [])}  # set comprehension
                                to_be_deleted = my_handles - updated_handles
                                for handle in to_be_deleted:
                                    state = self.context_states.handle.get_one(handle)
                                    self.context_states.remove_object_no_lock(state)
                        for state_container in updated_state_containers:
                            my_multi_key = multi_key(state_container)

                            if state_container.is_context_state:
                                old_state_container = my_multi_key.handle.get_one(state_container.Handle,
                                                                                  allow_none=True)
                            else:
                                old_state_container = my_multi_key.descriptor_handle.get_one(
                                    state_container.DescriptorHandle, allow_none=True)
                                if old_state_container is None:
                                    self._logger.error(  # noqa: PLE1205
                                        'process_incoming_descriptors: got update of state "{}" , '
                                        'but it did not exist in mdib!',
                                        state_container.DescriptorHandle)
                            if old_state_container is not None:
                                old_state_container.update_from_other_container(state_container)
                                my_multi_key.update_object(old_state_container)

                    elif modification_type == dmt.DELETE:
                        deleted_descriptor_containers = report_part.Descriptor
                        deleted_state_containers = report_part.State
                        for descriptor_container in deleted_descriptor_containers:
                            self._logger.debug(  # noqa: PLE1205
                                'process_incoming_descriptors: remove descriptor "{}" (parent="{}")',
                                descriptor_container.Handle, descriptor_container.parent_handle)
                            self.rm_descriptor_by_handle(
                                descriptor_container.Handle)
                            deleted_descriptor_by_handle[descriptor_container.Handle] = descriptor_container

                        for state_container in deleted_state_containers:
                            multi_key(state_container).remove_object_no_lock(state_container)
                    else:
                        msg = f'unknown modification type {modification_type} in description modification report'
                        raise ValueError(msg)

        finally:
            self.description_modifications = report  # update observable for complete report
            # update observables for every report part separately
            if new_descriptor_by_handle:
                self.new_descriptors_by_handle = new_descriptor_by_handle
            if updated_descriptor_by_handle:
                self.updated_descriptors_by_handle = updated_descriptor_by_handle
            if deleted_descriptor_by_handle:
                self.deleted_descriptor_by_handle = deleted_descriptor_by_handle

    def _has_new_state_usable_state_version(self,
                                            old_state_container: AbstractStateContainer,
                                            new_state_container: AbstractStateContainer,
                                            report_name: str) -> bool:
        """Compare state versions old vs new.

        :param old_state_container:
        :param new_state_container:
        :param report_name: used for logging
        :return: True if new state is ok for mdib , otherwise False.
        """
        diff = int(new_state_container.StateVersion) - int(old_state_container.StateVersion)
        # diff == 0 can happen if there is only a descriptor version update
        if diff == 1:  # this is the perfect version
            return True
        if diff > 1:
            self._logger.error('{}: missed {} states for state DescriptorHandle={} ({}->{})',  # noqa: PLE1205
                               report_name,
                               diff - 1, old_state_container.DescriptorHandle,
                               old_state_container.StateVersion, new_state_container.StateVersion)
            return True  # the new version is newer, therefore it can be added to mdib
        if diff < 0:
            self._logger.error(  # noqa: PLE1205
                '{}: reduced state version for state DescriptorHandle={} ({}->{}) ',
                report_name, old_state_container.DescriptorHandle,
                old_state_container.StateVersion, new_state_container.StateVersion)
            return False
        diffs = old_state_container.diff(new_state_container)  # compares all xml attributes
        if diffs:
            self._logger.error(  # noqa: PLE1205
                '{}: repeated state version {} for state {}, DescriptorHandle={}, but states have different data:{}',
                report_name, old_state_container.StateVersion, old_state_container.__class__.__name__,
                old_state_container.DescriptorHandle, diffs)
        return False



from __future__ import annotations

from enum import Enum
from typing import Union, Optional

from .addressing_types import EndpointReferenceType
from . import xml_structure as cp
from sdc11073.namespaces import default_ns_helper
from .basetypes import XMLTypeBase, ElementWithText


class DeviceRelationshipTypeURI(str, Enum):
    HOST = "http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01/host"


class DeviceMetadataDialectURI(str, Enum):
    THIS_MODEL = "http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01/ThisModel"
    THIS_DEVICE = "http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01/ThisDevice"
    RELATIONSHIP = "http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01/Relationship"


class DeviceEventingFilterDialectURI(str, Enum):
    ACTION = "http://docs.oasis-open.org/ws-dd/ns/dpws/2009/01/Action"


class Relationship:
    __slots__ = ('host', 'hosted')

    def __init__(self):
        self.host = None
        self.hosted = {}


class LocalizedStringTypeDict(dict):
    """This class represents LocalizedStringType elements. It is a dictionary of lang:string entries.
     If lang is None, the_string is the default string."""

    def add_localized_string(self, the_string: str, lang: Optional[str] = None) -> None:
        """
        Method for better readability of code
        :param the_string:
        :param lang:
        :return: None
        """
        self[lang] = the_string


class HostServiceType(XMLTypeBase):
    EndpointReference = cp.SubElementProperty(default_ns_helper.WSA.tag('EndpointReference'),
                                              value_class=EndpointReferenceType)
    Types = cp.NodeTextQNameListProperty(default_ns_helper.DPWS.tag('Types'))
    _props = ('EndpointReference', 'Types')


class HostedServiceType(XMLTypeBase):
    EndpointReference = cp.SubElementListProperty(default_ns_helper.WSA.tag('EndpointReference'),
                                                  value_class=EndpointReferenceType)
    Types = cp.NodeTextQNameListProperty(default_ns_helper.DPWS.tag('Types'))
    ServiceId = cp.AnyUriTextElement(default_ns_helper.DPWS.tag('ServiceId'))
    _props = ('EndpointReference', 'Types', 'ServiceId')


class LocalizedStringType(ElementWithText):
    lang = cp.StringAttributeProperty(default_ns_helper.XML.tag('lang'))
    _props = ('lang',)

    @classmethod
    def init(cls, text: str, lang: Optional[str] = None):
        """
        This class represents the LocalizedStringType in DPWS.
        :param text: the text
        :param lang: if given, the actual language
        :return:
        """
        instance = cls()
        instance.lang = lang
        instance.text = text
        return instance


class ThisDeviceType(XMLTypeBase):
    """
    This class represents "ThisDeviceType" in dpws schema.
    """
    FriendlyName = cp.SubElementListProperty(default_ns_helper.DPWS.tag('FriendlyName'),
                                             value_class=LocalizedStringType)
    FirmwareVersion = cp.NodeStringProperty(default_ns_helper.DPWS.tag('FirmwareVersion'), is_optional=True)
    SerialNumber = cp.NodeStringProperty(default_ns_helper.DPWS.tag('SerialNumber'), is_optional=True)
    _props = ('FriendlyName', 'FirmwareVersion', 'SerialNumber')

    def __init__(self, friendly_name: Union[str, LocalizedStringTypeDict, None] = None,
                 firmware_version: Optional[str] = None,
                 serial_number: Optional[str] = None):
        """
        This class represents "ThisDeviceType" in dpws schema.
        :param friendly_name: If argument is a string, it is considered to be the default name.
                              If argument is a dictionary, it is expected to be key=language, value=name.
                              None as key marks the default name.
        :param firmware_version: any string
        :param serial_number: any string
        """
        super().__init__()
        if isinstance(friendly_name, str):
            self.FriendlyName.append(LocalizedStringType.init(friendly_name))
        elif isinstance(friendly_name, LocalizedStringTypeDict):
            for lang, text in friendly_name.items():
                self.FriendlyName.append(LocalizedStringType.init(text, lang))
        self.FirmwareVersion = firmware_version
        self.SerialNumber = serial_number


class ThisModelType(XMLTypeBase):
    Manufacturer = cp.SubElementListProperty(default_ns_helper.DPWS.tag('Manufacturer'),
                                             value_class=LocalizedStringType)
    ManufacturerUrl = cp.NodeStringProperty(default_ns_helper.DPWS.tag('ManufacturerUrl'))
    ModelName = cp.SubElementListProperty(default_ns_helper.DPWS.tag('ModelName'),
                                          value_class=LocalizedStringType)
    ModelNumber = cp.NodeStringProperty(default_ns_helper.DPWS.tag('ModelNumber'), is_optional=True)
    ModelUrl = cp.NodeStringProperty(default_ns_helper.DPWS.tag('ModelUrl'), is_optional=True)
    PresentationUrl = cp.NodeStringProperty(default_ns_helper.DPWS.tag('PresentationUrl'), is_optional=True)
    _props = ('Manufacturer', 'ManufacturerUrl', 'ModelName', 'ModelNumber', 'ModelUrl', 'PresentationUrl')

    def __init__(self,
                 manufacturer: Union[str, LocalizedStringTypeDict, None] = None,
                 manufacturer_url: Optional[str] = None,
                 model_name: Union[str, LocalizedStringTypeDict, None] = None,
                 model_number: Optional[str] = None,
                 model_url: Optional[str] = None,
                 presentation_url: Optional[str] = None):
        super().__init__()
        if isinstance(manufacturer, str):
            self.Manufacturer.append(LocalizedStringType.init(manufacturer))
        elif isinstance(manufacturer, LocalizedStringTypeDict):
            for lang, text in manufacturer.items():
                self.Manufacturer.append(LocalizedStringType.init(text, lang))
        self.ManufacturerUrl = manufacturer_url
        if isinstance(model_name, str):
            self.ModelName.append(LocalizedStringType.init(model_name))
        elif isinstance(model_name, LocalizedStringTypeDict):
            for lang, text in model_name.items():
                self.ModelName.append(LocalizedStringType.init(text, lang))
        self.ModelNumber = model_number
        self.ModelUrl = model_url
        self.PresentationUrl = presentation_url




"""Implementation of products.

A product is a set of role providers that handle operations and other tasks.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from sdc11073 import loghelper

from .alarmprovider import AlertDelegateProvider, AlertPreCommitHandler, AlertSystemStateMaintainer
from .audiopauseprovider import AudioPauseProvider
from .clockprovider import GenericSDCClockProvider
from .componentprovider import GenericSetComponentStateOperationProvider
from .contextprovider import EnsembleContextProvider, LocationContextProvider
from .metricprovider import GenericMetricProvider
from .operationprovider import OperationProvider
from .patientcontextprovider import GenericPatientContextProvider

if TYPE_CHECKING:
    from sdc11073.mdib.descriptorcontainers import AbstractOperationDescriptorProtocol
    from sdc11073.mdib.mdibprotocol import ProviderMdibProtocol
    from sdc11073.mdib.transactionsprotocol import AnyTransactionManagerProtocol
    from sdc11073.provider.operations import OperationDefinitionBase
    from sdc11073.provider.sco import AbstractScoOperationsRegistry

    from .providerbase import OperationClassGetter


class ProviderRoleProtocol(Protocol):
    """A ProviderRole implements operation handlers and can run other jobs that the role requires.

    This Interface is expected by BaseProduct.
    """

    def stop(self):
        """Stop worker threads etc."""

    def init_operations(self, sco: AbstractScoOperationsRegistry):
        """Init instance.

        Method is called on start.
        """

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Return a callable for this operation or None.

        If a mdib already has operations defined, this method can connect a handler to a given operation descriptor.
        Use case: initialization from an existing mdib
        """

    def make_missing_operations(self, sco: AbstractScoOperationsRegistry) -> list[OperationDefinitionBase]:
        """Make_missing_operations is called after all existing operations from mdib have been registered.

        If a role provider needs to add operations beyond that, it can do it here.
        """

    def on_pre_commit(self, mdib: ProviderMdibProtocol, transaction: AnyTransactionManagerProtocol):
        """Manipulate operation (e.g. add more states)."""

    def on_post_commit(self, mdib: ProviderMdibProtocol, transaction: AnyTransactionManagerProtocol):
        """Implement actions after the transaction."""
        ...


class BaseProduct:
    """A Product is associated to a single sco.

    It provides the operation handlers for the operations in this sco.
    If a mdib contains multiple sco instances, there must be multiple Products.
    """

    def __init__(self,
                 mdib: ProviderMdibProtocol,
                 sco: AbstractScoOperationsRegistry,
                 log_prefix: str | None = None):
        """Create a product."""
        self._sco = sco
        self._mdib = mdib
        self._ordered_providers: list[ProviderRoleProtocol] = []  # order matters, first come, first served
        # start with most specific providers, end with most general ones
        self._logger = loghelper.get_logger_adapter(f'sdc.device.{self.__class__.__name__}', log_prefix)

    def _all_providers_sorted(self) -> list[ProviderRoleProtocol]:
        return self._ordered_providers

    def init_operations(self):
        """Register all actively provided operations."""
        sco_handle = self._sco.sco_descriptor_container.Handle
        self._logger.info('init_operations for sco %s.', sco_handle)

        for role_handler in self._all_providers_sorted():
            role_handler.init_operations(self._sco)

        self._register_existing_mdib_operations(self._sco)

        for role_handler in self._all_providers_sorted():
            operations = role_handler.make_missing_operations(self._sco)
            if operations:
                info = ', '.join([f'{op.OP_DESCR_QNAME.localname} {op.handle}' for op in operations])
                self._logger.info('role handler %s added operations to mdib: %s',
                                  role_handler.__class__.__name__, info)
            for operation in operations:
                self._sco.register_operation(operation)

        all_sco_operations = self._mdib.entities.by_parent_handle(self._sco.sco_descriptor_container.Handle)
        all_op_handles = [op.handle for op in all_sco_operations]
        all_not_registered_op_handles = [op_h for op_h in all_op_handles if
                                         self._sco.get_operation_by_handle(op_h) is None]

        if not all_op_handles:
            self._logger.info('sco %s has no operations in mdib.', sco_handle)
        elif all_not_registered_op_handles:
            self._logger.info('sco %s has operations without handler! handles = %r',
                              sco_handle, all_not_registered_op_handles)
        else:
            self._logger.info('sco %s: all operations have a handler.', sco_handle)
        self._mdib.pre_commit_handler = self._on_pre_commit
        self._mdib.post_commit_handler = self._on_post_commit

    def stop(self):
        """Stop all role providers."""
        for role_handler in self._all_providers_sorted():
            role_handler.stop()

    def make_operation_instance(self,
                                operation_descriptor_container: AbstractOperationDescriptorProtocol,
                                operation_cls_getter: OperationClassGetter) -> OperationDefinitionBase | None:
        """Call make_operation_instance of all role providers, until the first returns not None."""
        operation_target_handle = operation_descriptor_container.OperationTarget
        operation_target_entity = self._mdib.entities.by_handle(operation_target_handle)
        if operation_target_entity is None:
            # this operation is incomplete, the operation target does not exist. Registration not possible.
            self._logger.warning('Operation %s: target %s does not exist, will not register operation',
                                 operation_descriptor_container.Handle, operation_target_handle)
            return None
        for role_handler in self._all_providers_sorted():
            operation = role_handler.make_operation_instance(operation_descriptor_container, operation_cls_getter)
            if operation is not None:
                self._logger.debug('%s provided operation for %s',
                                   role_handler.__class__.__name__, operation_descriptor_container)
                return operation
            self._logger.debug('%s: no handler for %s', role_handler.__class__.__name__, operation_descriptor_container)
        return None

    def _register_existing_mdib_operations(self, sco: AbstractScoOperationsRegistry):
        operation_entities = self._mdib.entities.by_parent_handle(self._sco.sco_descriptor_container.Handle)
        for operation_entity in operation_entities:
            registered_op = sco.get_operation_by_handle(operation_entity.handle)
            if registered_op is None:
                self._logger.debug('found unregistered %s in mdib, handle=%s, code=%r target=%s',
                                   operation_entity.node_type.localname, operation_entity.handle,
                                   operation_entity.descriptor.Type,
                                   operation_entity.descriptor.OperationTarget)
                operation = self.make_operation_instance(operation_entity.descriptor,
                                                         sco.operation_cls_getter)
                if operation is not None:
                    sco.register_operation(operation)

    def _on_pre_commit(self, mdib: ProviderMdibProtocol, transaction: AnyTransactionManagerProtocol):
        for provider in self._all_providers_sorted():
            provider.on_pre_commit(mdib, transaction)

    def _on_post_commit(self, mdib: ProviderMdibProtocol, transaction: AnyTransactionManagerProtocol):
        for provider in self._all_providers_sorted():
            provider.on_post_commit(mdib, transaction)


class DefaultProduct(BaseProduct):
    """Default Product."""

    def __init__(self,
                 mdib: ProviderMdibProtocol,
                 sco: AbstractScoOperationsRegistry,
                 log_prefix: str | None = None):
        super().__init__(mdib, sco, log_prefix)
        self.metric_provider = GenericMetricProvider(mdib, log_prefix=log_prefix)  # needed in a test
        self._ordered_providers.extend([AudioPauseProvider(mdib, log_prefix=log_prefix),
                                        GenericSDCClockProvider(mdib, log_prefix=log_prefix),
                                        GenericPatientContextProvider(mdib, log_prefix=log_prefix),
                                        AlertDelegateProvider(mdib, log_prefix=log_prefix),
                                        AlertSystemStateMaintainer(mdib, log_prefix=log_prefix),
                                        AlertPreCommitHandler(mdib, log_prefix=log_prefix),
                                        self.metric_provider,
                                        OperationProvider(mdib, log_prefix=log_prefix),
                                        GenericSetComponentStateOperationProvider(mdib, log_prefix=log_prefix),
                                        ])


class ExtendedProduct(DefaultProduct):
    """Add EnsembleContextProvider and LocationContextProvider."""

    def __init__(self,
                 mdib: ProviderMdibProtocol,
                 sco: AbstractScoOperationsRegistry,
                 log_prefix: str | None = None):
        super().__init__(mdib, sco, log_prefix)
        self._ordered_providers.extend([EnsembleContextProvider(mdib, log_prefix=log_prefix),
                                        LocationContextProvider(mdib, log_prefix=log_prefix),
                                        ])



"""Using lxml based SoapClient."""
from __future__ import annotations

import copy
import functools
import logging
import ssl
import time
import traceback
import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from lxml import etree

import sdc11073.certloader
from sdc11073 import commlog, loghelper, network, xml_utils
from sdc11073 import observableproperties as properties
from sdc11073.definitions_base import ProtocolsRegistry
from sdc11073.dispatch import DispatchKey, MessageConverterMiddleware
from sdc11073.exceptions import ApiUsageError
from sdc11073.httpserver import compression
from sdc11073.httpserver.httpserverimpl import HttpServerThreadBase
from sdc11073.namespaces import EventingActions
from sdc11073.xml_types import eventing_types, mex_types
from sdc11073.xml_types.addressing_types import HeaderInformationBlock
from sdc11073.xml_types.dpws_types import DeviceEventingFilterDialectURI
from sdc11073.xml_types.wsd_types import ProbeMatchesType, ProbeType

from .components import default_sdc_consumer_components
from .request_handler_deferred import EmptyResponse

if TYPE_CHECKING:
    from collections.abc import Iterable

    from sdc11073.consumer.serviceclients.serviceclientbase import HostedServiceClient
    from sdc11073.consumer.subscription import ConsumerSubscriptionManagerProtocol
    from sdc11073.definitions_base import AbstractDataModel, BaseDefinitions
    from sdc11073.dispatch.request import RequestData
    from sdc11073.entity_mdib.entity_consumermdib import EntityConsumerMdib
    from sdc11073.mdib.consumermdib import ConsumerMdib
    from sdc11073.pysoap.msgfactory import MessageFactory
    from sdc11073.pysoap.msgreader import MessageReader, ReceivedMessage
    from sdc11073.pysoap.soapclient import SoapClientProtocol
    from sdc11073.wsdiscovery.service import Service
    from sdc11073.xml_types.mex_types import HostedServiceType

    from .components import SdcConsumerComponents
    from .subscription import ConsumerSubscription


class HostedServiceDescription:
    """HostedServiceDescription collects initial structural data from provider."""

    def __init__(self, service_id: str,  # noqa: PLR0913
                 endpoint_address: str,
                 msg_reader: MessageReader,
                 msg_factory: MessageFactory,
                 data_model: AbstractDataModel,
                 log_prefix: str = ''):
        self._endpoint_address = endpoint_address
        self.service_id = service_id
        self._msg_reader = msg_reader
        self.msg_factory = msg_factory
        self._data_model = data_model
        self.log_prefix = log_prefix
        self.meta_data: mex_types.Metadata | None = None
        self.wsdl_string = None
        self.wsdl_node: xml_utils.LxmlElement | None = None
        self._logger = loghelper.get_logger_adapter('sdc.client.hosted', log_prefix)
        self._url = urlparse(endpoint_address)
        self.services = {}

    def read_metadata(self, soap_client: SoapClientProtocol):
        """Read metadata from provider.

        This gives consumer information about offered services and its addresses.
        Members meta_data, wsdl_node and wsdl_string are filled with data from provider.
        """
        payload = mex_types.GetMetadata()
        inf = HeaderInformationBlock(action=payload.action, addr_to=self._endpoint_address)
        created_message = self.msg_factory.mk_soap_message(inf, payload=payload)
        message_data = soap_client.post_message_to(self._url.path,
                                                   created_message,
                                                   msg=f'<{self.service_id}> read_metadata')
        self.meta_data = mex_types.Metadata.from_node(message_data.p_msg.body_node)
        if self.meta_data.wsdl_location is not None:
            self._read_wsdl(soap_client, self.meta_data.wsdl_location)

    def _read_wsdl(self, soap_client: SoapClientProtocol, wsdl_url: str):
        parsed = urlparse(wsdl_url)
        actual_path = parsed.path + f'?{parsed.query}' if parsed.query else parsed.path
        self.wsdl_bytes = soap_client.get_from_url(actual_path, msg=f'{self.log_prefix}:getwsdl')
        try:
            wsdl_element_tree = self._msg_reader.read_wsdl(self.wsdl_bytes)
            self.wsdl_node = wsdl_element_tree.getroot()
            try:
                encoding = wsdl_element_tree.docinfo.encoding or 'UTF-8'
            except AttributeError:
                encoding = 'UTF-8'
            self.wsdl_string = self.wsdl_bytes.decode(encoding)
            logging.getLogger(commlog.WSDL).debug(self.wsdl_string)
        except etree.XMLSyntaxError as ex:
            self._logger.error(  # noqa: PLE1205 TRY400
                'could not read wsdl from {}: error={}, data=\n{}', actual_path, ex, self.wsdl_bytes)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} "{self.service_id}" endpoint = {self._endpoint_address}'


@dataclass(frozen=True)
class SubscriptionEndData:
    """When a subscription end message is received, sdc consumer writes a SubscriptionEndData instance to observable.

    By doing so the application can react to the message.
    """

    subscription: ConsumerSubscription
    request_data: RequestData


class _NotificationsSplitter:
    """Distributes incoming notification to the specific observables."""

    def __init__(self, sdc_consumer: SdcConsumer):
        self._sdc_consumer = sdc_consumer
        self._lookup = self._mk_lookup()

    def on_notification(self, message_data: ReceivedMessage):
        observable_name = self._lookup.get(message_data.action)
        if observable_name is None:  # pragma: no cover
            how_i_hate_this = f'unknown message {message_data.action}'
            raise ValueError(how_i_hate_this)
        setattr(self._sdc_consumer, observable_name, message_data)

    def _mk_lookup(self) -> dict[str, str]:
        actions = self._sdc_consumer.sdc_definitions.Actions
        return {
            actions.Waveform: 'waveform_report',
            actions.EpisodicMetricReport: 'episodic_metric_report',
            actions.EpisodicAlertReport: 'episodic_alert_report',
            actions.EpisodicComponentReport: 'episodic_component_report',
            actions.EpisodicOperationalStateReport: 'episodic_operational_state_report',
            actions.EpisodicContextReport: 'episodic_context_report',
            actions.PeriodicMetricReport: 'periodic_metric_report',
            actions.PeriodicAlertReport: 'periodic_alert_report',
            actions.PeriodicComponentReport: 'periodic_component_report',
            actions.PeriodicOperationalStateReport: 'periodic_operational_state_report',
            actions.PeriodicContextReport: 'periodic_context_report',
            actions.DescriptionModificationReport: 'description_modification_report',
            actions.OperationInvokedReport: 'operation_invoked_report',
            actions.SystemErrorReport: 'system_error_report',
        }


class SdcConsumer:
    """The SdcConsumer can be used with a known device location.

    The location is typically the result of a wsdiscovery process.
    This class expects that the BICEPS services are available in the device.
    What if not???? => raise exception in _discover_hosted_services.
    """

    subscription_status = properties.ObservableProperty({})
    # indicates whether all subscriptions have been subscribed to and are renewed successfully
    is_connected = properties.ObservableProperty(default_value=False)

    # observable properties for all notifications
    # all incoming Notifications can be observed in state_event_report ( as soap envelope)
    state_event_report = properties.ObservableProperty()

    # the following observables can be used to observe the incoming notifications by message type.
    # They contain only the body node of the notification, not the envelope
    waveform_report: ReceivedMessage = properties.ObservableProperty()
    episodic_metric_report: ReceivedMessage = properties.ObservableProperty()
    episodic_alert_report: ReceivedMessage = properties.ObservableProperty()
    episodic_component_report: ReceivedMessage = properties.ObservableProperty()
    episodic_operational_state_report: ReceivedMessage = properties.ObservableProperty()
    episodic_context_report: ReceivedMessage = properties.ObservableProperty()
    periodic_metric_report: ReceivedMessage = properties.ObservableProperty()
    periodic_alert_report: ReceivedMessage = properties.ObservableProperty()
    periodic_component_report: ReceivedMessage = properties.ObservableProperty()
    periodic_operational_state_report: ReceivedMessage = properties.ObservableProperty()
    periodic_context_report: ReceivedMessage = properties.ObservableProperty()
    description_modification_report: ReceivedMessage = properties.ObservableProperty()
    operation_invoked_report: ReceivedMessage = properties.ObservableProperty()
    subscription_end_data: ReceivedMessage = properties.ObservableProperty()
    system_error_report: ReceivedMessage = properties.ObservableProperty()

    SSL_CIPHERS = None  # None : use SSL default

    def __init__(self, device_location: str,  # noqa: PLR0913 PLR0915
                 sdc_definitions: type[BaseDefinitions],
                 ssl_context_container: sdc11073.certloader.SSLContextContainer | None,
                 epr: str | uuid.UUID | None = None,
                 validate: bool = True,
                 log_prefix: str = '',
                 default_components: SdcConsumerComponents | None = None,
                 specific_components: SdcConsumerComponents | None = None,
                 request_chunk_size: int = 0,
                 socket_timeout: int = 5,
                 force_ssl_connect: bool = False,
                 alternative_hostname: str | None = None,
                 ):
        """Construct a SdcConsumer.

        :param device_location: the XAddr location for meta data,
                                e.g. http://10.52.219.67:62616/72c08f50-74cc-11e0-8092-027599143341
        :param sdc_definitions: a class derived from BaseDefinitions
        :param epr: the path of this client in http server
        :param ssl_context_container: used for ssl connection to device and for own HTTP Server (notifications receiver)
        :param validate: bool
        :param log_prefix: a string used as prefix for logging
        :param specific_components: a SdcConsumerComponents instance or None
        :param request_chunk_size: if value > 0, message is split into chunks of this size
        :param socket_timeout: timeout for connections to provider
        :param force_ssl_connect: True: only accept ssl connections (requires a ssl_context_container)
                                  False: if ssl_context_container is provided, consumer first tries an
                                         encrypted connection, and if this raises an SSLError,
                                         it tries an unencrypted connection
        :param alternative_hostname: if supplied this hostname is used in xaddr, default is to use numerical
                                     ipv4 address (can be used to use full qualified hostname)
        """
        if not device_location.startswith('http'):
            raise ValueError('Invalid device_location, it must be match http(s)://<netloc> syntax')
        self.is_ssl_connection: bool | None
        if force_ssl_connect:
            if ssl_context_container is None:
                raise ValueError(
                    'Invalid combination of ssl_connect (True) and ssl_context_container (None) parameters')
            self.is_ssl_connection = True
        elif ssl_context_container is None:
            self.is_ssl_connection = False
        else:
            self.is_ssl_connection = None  # options allow both, needs to be decided when connecting

        self._device_location = device_location
        self.sdc_definitions = sdc_definitions
        if default_components is None:
            default_components = default_sdc_consumer_components
        self._components = copy.deepcopy(default_components)
        if specific_components is not None:
            self._components.merge(specific_components)

        self.subscription_status: dict[str, bool] = {}

        # available after start_all
        self._network_adapter: network.NetworkAdapter | None = None

        self.log_prefix = log_prefix
        self.request_chunk_size = request_chunk_size
        self._socket_timeout = socket_timeout
        self._logger = loghelper.get_logger_adapter('sdc.client', self.log_prefix)
        self.host_description: mex_types.Metadata | None = None
        self.hosted_services = {}  # lookup by service id
        self._validate = validate
        self._alternative_hostname = alternative_hostname

        try:
            self._logger.info('Using SSL is enabled. TLS 1.3 Support = {}', ssl.HAS_TLSv1_3)  # noqa: PLE1205
        except AttributeError:
            self._logger.info('Using SSL is enabled. TLS 1.3 is not supported')
        self._ssl_context_container = ssl_context_container
        self._epr = epr or uuid.uuid4()

        self._http_server = None
        self._is_internal_http_server = False

        self._logger.info('created {} for {}', self.__class__.__name__, self._device_location)  # noqa: PLE1205

        self._compression_methods = compression.CompressionHandler.available_encodings[:]
        self._subscription_mgr = None
        self.operations_manager = None
        self._service_clients = {}
        self._mdib = None
        self._soap_clients = {}  # all http connections that this client holds
        self.peer_certificate = None
        self.binary_peer_certificate = None
        self.all_subscribed = False
        # look for schemas added by services and components spec
        additional_schema_specs = set(self._components.additional_schema_specs)
        for handler_cls in self._components.service_handlers:
            additional_schema_specs.update(handler_cls.additional_namespaces)
        msg_reader_cls = self._components.msg_reader_class
        self.msg_reader = msg_reader_cls(self.sdc_definitions,
                                         list(additional_schema_specs),
                                         self._logger,
                                         validate=validate)

        msg_factory_cls = self._components.msg_factory_class
        self.msg_factory = msg_factory_cls(self.sdc_definitions,
                                           list(additional_schema_specs),
                                           self._logger,
                                           validate=validate)

        action_dispatcher_class = self._components.action_dispatcher_class
        self._services_dispatcher = action_dispatcher_class(log_prefix)

        self._notifications_splitter = _NotificationsSplitter(self)

        self._msg_converter = MessageConverterMiddleware(
            self.msg_reader, self.msg_factory, self._logger, self._services_dispatcher)

        # parameters of start_all call, will be set later in start_all
        self._not_subscribed_actions_param: Iterable[str] | None = None
        self._fixed_renew_interval_param: float | None = None
        self._shared_http_server_param: Any | None = None
        self._check_get_service_param: bool | None = None

    def set_mdib(self, mdib: ConsumerMdib | EntityConsumerMdib | None):
        """SdcConsumer sometimes must know the mdib data (e.g. Set service, activate method)."""
        if mdib is not None and self._mdib is not None:
            raise ApiUsageError('SdcConsumer has already an registered mdib')
        self._mdib = mdib
        if self.client('Set') is not None:
            self.client('Set').register_mdib(mdib)
        if self.client('Context') is not None:
            self.client('Context').register_mdib(mdib)

    @property
    def mdib(self) -> ConsumerMdib | None:
        """Return associated mdib."""
        return self._mdib

    @property
    def network_adapter(self) -> network.NetworkAdapter | None:
        """The network adapter used by this consumer."""
        return self._network_adapter

    @property
    def _epr_urn(self) -> str:
        """Return end point reference, e.g 'urn:uuid:8c26f673-fdbf-4380-b5ad-9e2454a65b6b'."""
        try:
            return self._epr.urn
        except AttributeError:
            return self._epr

    @property
    def path_prefix(self) -> str:
        """Return http path prefix of service e.g '8c26f673fdbf4380b5ad9e2454a65b6b'."""
        try:
            return self._epr.hex
        except AttributeError:
            return self._epr

    @property
    def base_url(self) -> str:
        """Return the base url of consumer.

        Replace servers ip address with own ip address (server might have 0.0.0.0).
        """
        if self._http_server is None:
            return ''
        p = urlparse(self._http_server.base_url)
        tmp = f'{p.scheme}://{self._alternative_hostname or self._network_adapter.ip}:{p.port}{p.path}'
        sep = '' if tmp.endswith('/') else '/'
        return f'{tmp}{sep}{self.path_prefix}/'

    def mk_subscription(self, dpws_hosted: HostedServiceType,
                        filter_type: eventing_types.FilterType,
                        actions: Iterable[DispatchKey]) -> ConsumerSubscription:
        """Create a subscription object and registers it in dispatcher.

        :param dpws_hosted: proxy for the hosted service that provides the events we want to subscribe to
                           This is the target for all subscribe/unsubscribe ... messages
        :param filter_type: the filter that is sent to device
        :param actions: a list of DispatchKey that this subscription shall handle.
        :return: a subscription object.
        """
        subscription = self._subscription_mgr.mk_subscription(dpws_hosted, filter_type)

        def update_subscription_status(subscription_filter: str, status: bool):
            subscription_status = dict(self.subscription_status)
            subscription_status[subscription_filter] = status
            self.subscription_status = subscription_status  # trigger observable if status has changed

        properties.strongbind(subscription, is_subscribed=functools.partial(update_subscription_status,
                                                                            filter_type.text))

        # direct subscribed notifications to this subscription
        for action in actions:
            self._services_dispatcher.register_post_handler(action, subscription.on_notification)
        return subscription

    def do_subscribe(self, dpws_hosted: HostedServiceType,  # noqa: PLR0913
                     filter_type: eventing_types.FilterType,
                     actions: Iterable[DispatchKey],
                     expire_minutes: int = 60,
                     any_elements: list[xml_utils.LxmlElement] | None = None,
                     any_attributes: dict | None = None) -> ConsumerSubscription:
        """Send subscribe request to provider.

        :param dpws_hosted: proxy for the hosted service that provides the events we want to subscribe to
                           This is the target for all subscribe/unsubscribe ... messages
        :param filter_type: the filter that is sent to device
        :param actions: a list of DispatchKeys that this subscription shall handle
        :param expire_minutes: defaults to 1 hour
        :param any_elements: optional list of lxml elements
        :param any_attributes: optional dictionary of name:str - value:str pairs
        :return: a subscription object that has callback already registered.
        """
        subscription = self.mk_subscription(dpws_hosted, filter_type, actions)
        properties.bind(subscription, notification_data=self._on_notification)
        subscription.subscribe(expire_minutes, any_elements, any_attributes)
        return subscription

    def client(self, port_type_name: str) -> HostedServiceClient | None:
        """Return the client for the given port type name.

        WDP and SDC use different port type names, e.g. WPF="Get", SDC="GetService".
        If the port type is not found directly, it tries also with or without "Service" in name.
        :param port_type_name: string, e.g "Get", or "GetService", ...
        """
        client = self._service_clients.get(port_type_name)
        if client is None and port_type_name.endswith('Service'):
            client = self._service_clients.get(port_type_name[:-7])
        if client is None and not port_type_name.endswith('Service'):
            client = self._service_clients.get(port_type_name + "Service")
        return client

    @property
    def get_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('GetService')

    @property
    def set_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('SetService')

    @property
    def description_event_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('DescriptionEventService')

    @property
    def state_event_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('StateEventService')

    @property
    def context_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('ContextService')

    @property
    def waveform_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('Waveform')

    @property
    def containment_tree_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('ContainmentTreeService')

    @property
    def archive_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('ArchiveService')

    @property
    def localization_service_client(self) -> HostedServiceClient:
        """Return a HostedServiceClient."""
        return self.client('LocalizationService')

    @property
    def subscription_mgr(self) -> ConsumerSubscriptionManagerProtocol:
        """Return the subscription manager."""
        return self._subscription_mgr

    def start_all(self, not_subscribed_actions: Iterable[str] | None = None,  #noqa: C901 PLR0915
                  fixed_renew_interval: float | None = None,
                  shared_http_server: Any | None = None,
                  check_get_service: bool = True) -> None:
        """Start background threads, read metadata from device, instantiate detected port type clients and subscribe.

        :param not_subscribed_actions: a list of pmtypes.Actions elements or None. if None, everything is subscribed.
        :param fixed_renew_interval: an interval in seconds or None
                    if None, renew is sent when remaining time <= 50% of granted time
                    if set, subscription renew is sent in this interval.
        :param shared_http_server: if provided, use this http server, else client creates its own.
        :param check_get_service: if True (default) it checks that a GetService is detected,
               which is the minimal requirement for a sdc provider.
        :return: None
        """
        self._not_subscribed_actions_param = not_subscribed_actions
        self._fixed_renew_interval_param = fixed_renew_interval
        self._shared_http_server_param = shared_http_server
        self._check_get_service_param = check_get_service
        self._logger.debug('connecting to %s', self._device_location)
        self._connect()
        self._logger.debug('reading meta data from %s', self._device_location)
        self.host_description = self._get_metadata()

        # now query also metadata of hosted services
        self._mk_hosted_services(self.host_description)
        self._logger.debug('Services: {}', self._service_clients.keys())  # noqa: PLE1205

        used_ip = self.get_soap_client(self._device_location).sock_name[0]
        self._network_adapter = network.get_adapter_containing_ip(used_ip)
        self._logger.info('SdcConsumer for {} uses network adapter {}',  # noqa: PLE1205
                          self._device_location,
                          self._network_adapter)

        # only GetService is mandatory!!!
        if check_get_service and self.get_service_client is None:  # pragma: no cover
            msg = f'GetService not detected! found services = {list(self._service_clients.keys())}'
            raise RuntimeError(msg)

        self._start_event_sink(shared_http_server)

        # start subscription manager
        subscription_manager_class = self._components.subscription_manager_class
        self._subscription_mgr = subscription_manager_class(self.msg_reader,
                                                            self.msg_factory,
                                                            self.sdc_definitions.data_model,
                                                            self.get_soap_client,
                                                            self.base_url,
                                                            log_prefix=self.log_prefix,
                                                            fixed_renew_interval=fixed_renew_interval)
        self._subscription_mgr.start()

        # flag 'self.all_subscribed' tells mdib that mdib state versions shall not have any gaps
        # => log warnings for missing versions
        self.all_subscribed = True
        not_subscribed_actions_set = set() if not_subscribed_actions is None else set(not_subscribed_actions)
        if not_subscribed_actions:
            not_subscribed_episodic_actions = [a for a in not_subscribed_actions
                                               if ("Episodic" in a or "DescriptionModificationReport" in a)]
            if not_subscribed_episodic_actions:
                self.all_subscribed = False

        # start operationInvoked subscription and tell all
        operations_manager_class = self._components.operations_manager_class
        self.operations_manager = operations_manager_class(self.msg_reader, self.log_prefix)
        properties.bind(self, operation_invoked_report=self.operations_manager.on_operation_invoked_report)
        for client in self._service_clients.values():
            client.set_operations_manager(self.operations_manager)

        # start all subscriptions
        # group subscriptions per hosted service
        for dpws_hosted in self.host_description.relationship.Hosted:
            available_actions: list[DispatchKey] = []
            if dpws_hosted.Types is not None:
                for port_type_qname in dpws_hosted.Types:
                    client = self.client(port_type_qname.localname)
                    if client is not None:
                        available_actions.extend(client.get_available_subscriptions())
            if len(available_actions) > 0:
                subscribe_actions = {a for a in available_actions if a.action not in not_subscribed_actions_set}
                if len(subscribe_actions) > 0:
                    filter_type = eventing_types.FilterType()
                    filter_type.text = ' '.join(x.action for x in subscribe_actions)
                    filter_type.Dialect = DeviceEventingFilterDialectURI.ACTION
                    try:
                        self.do_subscribe(dpws_hosted, filter_type, subscribe_actions)
                    except Exception:  # noqa: BLE001
                        self.all_subscribed = False  # => don't log errors when mdib versions are missing
                        self._logger.error('start_all: could not subscribe: error = {}, actions= {}',  # noqa: PLE1205 TRY400
                                           traceback.format_exc(), subscribe_actions)

        def _update_is_connected(subscription_status: dict[str, bool]):
            self.is_connected = all(subscription_status.values()) and any(subscription_status)

        properties.strongbind(self, subscription_status=_update_is_connected)
        _update_is_connected(self.subscription_status)

        # register callback for end of subscription
        self._services_dispatcher.register_post_handler(
            DispatchKey(EventingActions.SubscriptionEnd,
                        self.sdc_definitions.data_model.ns_helper.WSE.tag('SubscriptionEnd')),
            self._on_subscription_end)

    def stop_all(self, unsubscribe: bool = True):
        """Stop all threads, optionally unsubscribe."""
        if self._subscription_mgr is not None:
            if unsubscribe:
                self._subscription_mgr.unsubscribe_all()
            self._subscription_mgr.stop()
        self.set_mdib(None)

        for client in self._soap_clients.values():
            client.close()
        self._soap_clients = {}
        self._stop_event_sink()

    def restart(self):
        """Forget existing data and restart from the beginning."""
        mdib = self._mdib  # keep existing mdib connection
        self.stop_all()  # with unsubscribe
        # start with the same parameters as initially
        self.start_all(self._not_subscribed_actions_param,
                       self._fixed_renew_interval_param,
                       self._shared_http_server_param,
                       self._check_get_service_param)
        self.set_mdib(mdib)

    def set_used_compression(self, *compression_methods: str):
        """Use only one of these compression methods."""
        del self._compression_methods[:]
        self._compression_methods.extend(compression_methods)

    def _connect(self):
        soap_client = self.get_soap_client(self._device_location)
        if self.is_ssl_connection is not None:
            # decision was already made in constructor
            soap_client.connect()
        else:
            try:
                soap_client.connect()
                self.is_ssl_connection = True
            except ssl.SSLError:
                # could not connect with ssl, try without it
                soap_client.close()
                self._forget_soap_client(soap_client)
                self.is_ssl_connection = False
                soap_client = self.get_soap_client(self._device_location)
                # if this also fails, something else is wrong and error needs handling on application level.
                soap_client.connect()
        if self.is_ssl_connection:
            sock = soap_client.sock
            self.peer_certificate = sock.getpeercert(binary_form=False)
            self.binary_peer_certificate = sock.getpeercert(binary_form=True)  # in case the application needs it...
            self._logger.info('Peer Certificate: {}', self.peer_certificate)  # noqa: PLE1205

    def _get_metadata(self) -> mex_types.Metadata:
        _url = urlparse(self._device_location)
        soap_client = self.get_soap_client(self._device_location)
        nsh = self.sdc_definitions.data_model.ns_helper
        inf = HeaderInformationBlock(action=f'{nsh.WXF.namespace}/Get',
                                     addr_to=self._device_location)
        message = self.msg_factory.mk_soap_message_etree_payload(inf, payload_element=None)

        received_message_data = soap_client.post_message_to(_url.path, message, msg='getMetadata')
        return mex_types.Metadata.from_node(received_message_data.p_msg.body_node)

    def send_probe(self) -> ProbeMatchesType:
        """Send Probe directly to provider."""
        _url = urlparse(self._device_location)
        wsc = self.get_soap_client(self._device_location)
        probe = ProbeType()
        inf = HeaderInformationBlock(action=probe.action,
                                     addr_to=self._device_location)

        message = self.msg_factory.mk_soap_message(inf, payload=probe)
        received_message_data = wsc.post_message_to(_url.path, message, msg='Probe')
        return ProbeMatchesType.from_node(received_message_data.p_msg.msg_node)

    def get_soap_client(self, address: str) -> SoapClientProtocol:
        """Return the soap client for address.

        Method creates a new soap client if needed and considers self.is_ssl_connection value.
        """
        _url = urlparse(address)
        use_ssl = self.is_ssl_connection is not False  # if is_ssl_connection is still None, default to use_ssl = True
        key = (use_ssl, _url.netloc)
        soap_client = self._soap_clients.get(key)
        if soap_client is None:
            soap_client = self._mk_soap_client(use_ssl, _url.netloc)
            self._soap_clients[key] = soap_client
        return soap_client

    def _forget_soap_client(self, soap_client: SoapClientProtocol):
        for key, value in self._soap_clients.items():
            if value is soap_client:
                del self._soap_clients[key]
                return

    def _mk_soap_client(self, use_ssl: bool, netloc: str) -> SoapClientProtocol:
        _ssl_context = self._ssl_context_container.client_context if use_ssl else None
        cls = self._components.soap_client_class
        return cls(netloc,
                   self._socket_timeout,
                   loghelper.get_logger_adapter('sdc.client.soap', self.log_prefix),
                   ssl_context=_ssl_context,
                   sdc_definitions=self.sdc_definitions,
                   msg_reader=self.msg_reader,
                   supported_encodings=self._compression_methods,
                   chunk_size=self.request_chunk_size)

    def _mk_hosted_services(self, host_description: mex_types.Metadata):
        for hosted in host_description.relationship.Hosted:
            address = hosted.EndpointReference[0].Address
            soap_client = self.get_soap_client(address)
            h_descr = HostedServiceDescription(
                hosted.ServiceId, address,
                self.msg_reader, self.msg_factory, self.sdc_definitions.data_model, self.log_prefix)
            self.hosted_services[hosted.ServiceId] = h_descr
            h_descr.read_metadata(soap_client)
            for port_type in hosted.Types:
                hosted_service_client = self._mk_hosted_service_client(port_type,
                                                                       soap_client,
                                                                       hosted)
                if hosted_service_client is not None:
                    self._service_clients[port_type.localname] = hosted_service_client
                    h_descr.services[port_type.localname] = hosted_service_client
                else:
                    self._logger.warning('Unknown port type {}', str(port_type))  # noqa: PLE1205

    def _mk_hosted_service_client(self, port_type: str,
                                  soap_client: SoapClientProtocol,
                                  hosted: HostedServiceType) -> HostedServiceClient | None:
        for cls in self._components.service_handlers:
            if cls.port_type_name == port_type:
                return cls(self, soap_client, hosted, port_type)
        return None

    def _start_event_sink(self, shared_http_server: Any):
        if shared_http_server is None:
            self._is_internal_http_server = True
            ssl_context_container = self._ssl_context_container if self.is_ssl_connection else None
            logger = loghelper.get_logger_adapter('sdc.client.notif_dispatch', self.log_prefix)
            self._http_server = HttpServerThreadBase(
                str(self._network_adapter.ip),
                ssl_context_container.server_context if ssl_context_container else None,
                logger=logger,
                supported_encodings=self._compression_methods,
            )
            self._http_server.start()
            self._http_server.started_evt.wait(timeout=5)
            # it sometimes still happens that http server is not completely started without waiting.
            # find better solution, see issue #320
            time.sleep(1)
            self._logger.info('serving EventSink on {}', self._http_server.base_url)  # noqa: PLE1205
        else:
            self._http_server = shared_http_server
        # register own epr in http server
        self._http_server.dispatcher.register_instance(self.path_prefix, self._msg_converter)

    def _stop_event_sink(self):
        if self._is_internal_http_server and self._http_server is not None:
            self._http_server.stop()

    def _on_notification(self, message_data: ReceivedMessage):
        self.state_event_report = message_data  # update observable
        self._notifications_splitter.on_notification(message_data)

    def _on_subscription_end(self, request_data: RequestData) -> EmptyResponse:
        subscription = self._subscription_mgr.on_subscription_end(request_data)  # subscription can be None
        self.subscription_end_data = SubscriptionEndData(subscription, request_data)
        return EmptyResponse()

    def __str__(self) -> str:
        return (f'SdcConsumer to {self.host_description.this_device} {self.host_description.this_model} '
                f'on {self._device_location}')

    @classmethod
    def from_wsd_service(cls, wsd_service: Service,  # noqa: PLR0913
                         ssl_context_container: sdc11073.certloader.SSLContextContainer | None,
                         validate: bool = True,
                         log_prefix: str = '',
                         default_components: SdcConsumerComponents | None = None,
                         specific_components: SdcConsumerComponents | None = None) -> SdcConsumer:
        """Construct a SdcConsumer from a Service.

        :param wsd_service: a wsdiscovery.Service instance
        :param ssl_context_container: a ssl context or None
        :param validate: bool
        :param log_prefix: a string
        :param default_components: a SdcConsumerComponents instance or None
        :param specific_components: a SdcConsumerComponents instance or None
        :return:
        """
        device_locations = wsd_service.x_addrs
        if not device_locations:  # pragma: no cover
            msg = f'discovered Service has no address!{wsd_service}'
            raise RuntimeError(msg)
        device_location = device_locations[0]
        for sdc_definition in ProtocolsRegistry.protocols:
            if sdc_definition.types_match(wsd_service.types):
                return cls(device_location, sdc_definition, ssl_context_container, validate=validate,
                           log_prefix=log_prefix,
                           default_components=default_components, specific_components=specific_components)
        raise RuntimeError('no matching protocol definition found for this service!')



from sdc11073.dispatch import DispatchKey
from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types import msg_qnames
from sdc11073.xml_types.actions import Actions

from .serviceclientbase import HostedServiceClient


class DescriptionEventClient(HostedServiceClient):
    """Client for DescriptionEventService."""

    port_type_name = PrefixesEnum.SDC.tag('DescriptionEventService')
    notifications = (DispatchKey(Actions.DescriptionModificationReport, msg_qnames.DescriptionModificationReport),)



from __future__ import annotations

import enum
import inspect
import traceback
from math import isclose
from typing import TYPE_CHECKING

from lxml import etree

from .xml_structure import NodeStringProperty, NodeTextListProperty

if TYPE_CHECKING:
    from sdc11073 import xml_utils


class StringEnum(str, enum.Enum):

    def __str__(self):
        return str(self.value)


class XMLTypeBase:
    """ Base class that it used to declare XML data types. It supports nesting of data and inheritance.
    It uses xml_structure elements to declare the members.
    Because order matters in XML, the _props member is needed that lists all members that
    represent XML data in the correct order.
    Usage:
    - object creation: All derived classes have a constructor without arguments.
    - initializing from XML: class method 'from_node'
    -
    """

    def __init__(self):
        for _, prop in self.sorted_container_properties():
            prop.init_instance_data(self)

    def as_etree_node(self, q_name: etree.QName, ns_map: dict, parent_node: etree.Element | None = None):
        if parent_node is not None:
            node = etree.SubElement(parent_node, q_name, nsmap=ns_map)
        else:
            node = etree.Element(q_name, nsmap=ns_map)
        self.update_node(node)
        return node

    def update_node(self, node: xml_utils.LxmlElement):
        for prop_name, prop in self.sorted_container_properties():
            try:
                prop.update_xml_value(self, node)
            except Exception as ex:
                # re-raise with some information about the data
                raise ValueError(
                    f'In {self.__class__.__name__}.{prop_name}, {prop!s} could not update: {traceback.format_exc()}') from ex

    def update_from_node(self, node: xml_utils.LxmlElement):
        for dummy, prop in self.sorted_container_properties():
            prop.update_from_node(self, node)

    def sorted_container_properties(self):
        """
        @return: a list of (name, object) tuples of all GenericProperties ( and subclasses)
        list is created based on _props lists of classes
        """
        ret = []
        classes = inspect.getmro(self.__class__)
        for cls in reversed(classes):
            try:
                names = cls.__dict__['_props']  # this checks only current class, not parent
            except (AttributeError, KeyError):
                continue
            for name in names:
                obj = getattr(cls, name)
                if obj is not None:
                    ret.append((name, obj))
        return ret

    def __eq__(self, other):
        """ compares all properties"""
        try:
            for name, dummy in self.sorted_container_properties():
                my_value = getattr(self, name)
                other_value = getattr(other, name)
                if my_value == other_value:
                    continue
                if (isinstance(my_value, float) or isinstance(other_value, float)) and isclose(my_value, other_value):
                    continue  # float compare (almost equal)
                return False
            return True
        except (TypeError, AttributeError):
            return False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return f'{self.__class__.__name__}({self.sorted_container_properties()})'

    @classmethod
    def from_node(cls, node):
        """ default from_node Constructor that provides no arguments for class __init__"""
        obj = cls()
        obj.update_from_node(node)
        return obj

    @classmethod
    def value_class_from_node(cls, _):
        return cls


class ElementWithText(XMLTypeBase):
    """An Element with text. It is different form NodeTextProperty in two aspects:
    - access to text via "text" member, it is not the property value itself.
    - It can be extended with Attributes
    """
    NODETYPE = None
    text: str = NodeStringProperty()  # this is the text of the node. Here attribute is lower case!
    _props = ('text',)

    def __init__(self, text=None):
        super().__init__()
        self.text = text


class ElementWithTextList(XMLTypeBase):
    """An Element with text, which is alist of words(string without whitespace).
    """
    # this is the text list of the node. Here attribute is lower case!
    text: list[str] = NodeTextListProperty(sub_element_name=None,
                                           value_class=str)
    _props = ('text',)


class MessageType(XMLTypeBase):
    """This is the base for all classes that are used as the body of a soap envelope.
    All derived classes must set these values.
    NODETYPE defines the qualified name of the Element, action is used for the action element
    in the soap header."""
    NODETYPE = None
    action = None
    additional_namespaces = ()  # derived class list namespaces other than PM and MSG



"""The module contains extensions to the functionality of the ProviderMdib."""
from __future__ import annotations

import time
import uuid
from typing import TYPE_CHECKING, Any

from sdc11073.xml_types.pm_types import RetrievabilityMethod

if TYPE_CHECKING:
    from sdc11073.location import SdcLocation
    from sdc11073.xml_types.pm_types import (
        CodedValue,
        InstanceIdentifier,
        MetricAvailability,
        MetricCategory,
        SafetyClassification,
    )

    from .descriptorcontainers import AbstractDescriptorProtocol
    from .entityprotocol import MultiStateEntityProtocol
    from .providermdib import ProviderMdib
    from .statecontainers import AbstractStateProtocol

class ProviderMdibMethods:
    """Extra methods for provider mdib tht are not core functionality."""

    def __init__(self, provider_mdib: ProviderMdib):
        self._mdib = provider_mdib
        self.descriptor_factory = DescriptorFactory(provider_mdib)
        self.default_validators = (provider_mdib.data_model.pm_types.InstanceIdentifier(
            root='rootWithNoMeaning', extension_string='System'),)

    def ensure_location_context_descriptor(self):
        """Create a LocationContextDescriptor if there is none in mdib."""
        mdib = self._mdib
        pm = mdib.data_model.pm_names
        system_context_descriptors = mdib.descriptions.NODETYPE.get(pm.SystemContextDescriptor, [])
        location_context_descriptors = mdib.descriptions.NODETYPE.get(pm.LocationContextDescriptor, [])

        for system_context_descriptor in system_context_descriptors:
            child_location_descriptors = [d for d in location_context_descriptors
                                          if d.parent_handle == system_context_descriptor.Handle
                                          and pm.LocationContextDescriptor == d.NODETYPE]
            if not child_location_descriptors:
                descr_cls = mdib.data_model.get_descriptor_container_class(pm.LocationContextDescriptor)
                descr_container = descr_cls(handle=uuid.uuid4().hex, parent_handle=system_context_descriptor.Handle)
                descr_container.SafetyClassification = mdib.data_model.pm_types.SafetyClassification.INF
                mdib.descriptions.add_object(descr_container)

    def ensure_patient_context_descriptor(self):
        """Create a PatientContextDescriptor if there is none in mdib."""
        mdib = self._mdib
        pm = mdib.data_model.pm_names
        system_context_descriptors = mdib.descriptions.NODETYPE.get(pm.SystemContextDescriptor, [])
        patient_context_descriptors = mdib.descriptions.NODETYPE.get(pm.PatientContextDescriptor, [])

        for system_context_descriptor in system_context_descriptors:
            child_location_descriptors = [d for d in patient_context_descriptors
                                          if d.parent_handle == system_context_descriptor.Handle
                                          and pm.PatientContextDescriptor == d.NODETYPE]
            if not child_location_descriptors:
                descr_cls = mdib.data_model.get_descriptor_container_class(pm.PatientContextDescriptor)
                descr_container = descr_cls(handle=uuid.uuid4().hex, parent_handle=system_context_descriptor.Handle)
                descr_container.SafetyClassification = mdib.data_model.pm_types.SafetyClassification.INF
                mdib.descriptions.add_object(descr_container)

    def set_location(self, sdc_location: SdcLocation,
                     validators: list[InstanceIdentifier] | None = None,
                     location_context_descriptor_handle: str | None = None):
        """Create a location context state. The new state will be the associated state.

        This method updates only the mdib data!
        Use the SdcProvider.set_location method if you want to publish the address on the network.
        :param sdc_location: a sdc11073.location.SdcLocation instance
        :param validators: a list of InstanceIdentifier objects or None
               If None, self.default_validators is used.
        :param location_context_descriptor_handle: Only needed if the mdib contains more than one
               LocationContextDescriptor. Then this defines the descriptor for which a new LocationContextState
               shall be created.
        """
        mdib = self._mdib
        pm = mdib.data_model.pm_names

        if location_context_descriptor_handle is None:
            # assume there is only one descriptor in mdib, user has not provided a handle.
            descriptor_container = mdib.descriptions.NODETYPE.get_one(pm.LocationContextDescriptor)
        else:
            descriptor_container = mdib.descriptions.handle.get_one(location_context_descriptor_handle)

        with mdib.context_state_transaction() as mgr:
            mgr.disassociate_all(descriptor_container.Handle)

            new_location = mgr.mk_context_state(descriptor_container.Handle, set_associated=True)
            new_location.update_from_sdc_location(sdc_location)
            if validators is None:
                new_location.Validator = self.default_validators
            else:
                new_location.Validator = validators

    def mk_state_containers_for_all_descriptors(self):
        """Create a state container for every descriptor that is missing a state in mdib.

        The model requires that there is a state for every descriptor (exception: multi-states)
        """
        mdib = self._mdib
        pm = mdib.data_model.pm_names
        for descr in mdib.descriptions.objects:
            if descr.Handle not in mdib.states.descriptor_handle \
                    and descr.Handle not in mdib.context_states.descriptor_handle:
                state_cls = mdib.data_model.get_state_class_for_descriptor(descr)
                if state_cls.is_multi_state:
                    pass  # nothing to do, it is allowed to have no state
                else:
                    state = state_cls(descr)
                    # add some initial values where needed
                    if state.is_alert_condition:
                        state.DeterminationTime = time.time()
                    elif state.NODETYPE == pm.AlertSystemState:  # noqa: SIM300
                        state.LastSelfCheck = time.time()
                        state.SelfCheckCount = 1
                    elif state.NODETYPE == pm.ClockState:  # noqa: SIM300
                        state.LastSet = time.time()
                    if mdib.current_transaction is not None:
                        mdib.current_transaction.add_state(state)
                    else:
                        mdib.states.add_object(state)

    def set_states_initial_values(self):
        """Set all states to defined starting conditions.

        This method is ment to be called directly after the mdib was loaded and before the provider is published
        on the network.
        It changes values only internally in the mdib, no notifications are sent!

        """
        pm_names = self._mdib.data_model.pm_names
        pm_types = self._mdib.data_model.pm_types

        for state in self._mdib.states.objects:
            descriptor = self._mdib.descriptions.handle.get_one(state.DescriptorHandle)
            if pm_names.AlertSystemDescriptor == descriptor.NODETYPE:
                # alert systems are active
                state.ActivationState = pm_types.AlertActivation.ON
                state.SystemSignalActivation.append(
                    pm_types.SystemSignalActivation(manifestation=pm_types.AlertSignalManifestation.AUD,
                                                    state=pm_types.AlertActivation.ON))
            elif descriptor.is_alert_condition_descriptor:
                # alert conditions are active, but not present
                state.ActivationState = pm_types.AlertActivation.ON
                state.Presence = False
            elif descriptor.is_alert_signal_descriptor:
                # alert signals are not present, and delegable signals are also not active
                if descriptor.SignalDelegationSupported:
                    state.Location = pm_types.AlertSignalPrimaryLocation.REMOTE
                    state.ActivationState = pm_types.AlertActivation.OFF
                    state.Presence = pm_types.AlertSignalPresence.OFF
                else:
                    state.ActivationState = pm_types.AlertActivation.ON
                    state.Presence = pm_types.AlertSignalPresence.OFF
            elif descriptor.is_component_descriptor:
                # all components are active
                state.ActivationState = pm_types.ComponentActivation.ON
            elif descriptor.is_operational_descriptor:
                # all operations are enabled
                state.OperatingMode = pm_types.OperatingMode.ENABLED

    def update_retrievability_lists(self):
        """Update internal lists, based on current mdib descriptors."""
        mdib = self._mdib
        with mdib.mdib_lock:
            del mdib._retrievability_episodic[:]  # noqa: SLF001
            mdib.retrievability_periodic.clear()
            for descr in mdib.descriptions.objects:
                for r in descr.get_retrievability():
                    for r_by in r.By:
                        if r_by.Method == RetrievabilityMethod.EPISODIC:
                            mdib._retrievability_episodic.append(descr.Handle)  # noqa: SLF001
                        elif r_by.Method == RetrievabilityMethod.PERIODIC:
                            period_float = r_by.UpdatePeriod
                            period_ms = int(period_float * 1000.0)
                            mdib.retrievability_periodic[period_ms].append(descr.Handle)

    def set_all_source_mds(self):
        """Set the source mds member of all descriptors.

        This makes handling of SourceMds separation in messages less demanding on CPU.
        """
        pm_qnames = self._mdib.data_model.pm_names
        all_mds_descriptors = self._mdib.descriptions.NODETYPE.get(pm_qnames.MdsDescriptor)
        for mds_descriptor in all_mds_descriptors:
            for descr in self._mdib.get_all_descriptors_in_subtree(mds_descriptor):
                descr.set_source_mds(mds_descriptor.Handle)

    def get_mds_descriptor(self, container: AbstractDescriptorProtocol | AbstractStateProtocol) \
            -> AbstractDescriptorProtocol | None:
        """Get the parent mds descriptor for a given descriptor or state."""
        tmp = container
        if tmp.is_state_container:
            tmp = self._mdib.descriptions.handle.get_one(tmp.DescriptorHandle)
        mds_descr = None
        expected_type = self._mdib.data_model.pm_names.MdsDescriptor
        while mds_descr is None:
            if tmp.NODETYPE == expected_type:  # noqa: SIM300
                return tmp
            parent_handle = tmp.parent_handle
            tmp = self._mdib.descriptions.handle.get_one(parent_handle, allow_none=True)
            if tmp is None:
                if self._mdib.current_transaction:
                    tmp = self._mdib.current_transaction.actual_descriptor(parent_handle)
            if tmp is None:
                msg = f'could not find mds descriptor for handle {container.Handle}'
                raise KeyError(msg)
        return None

    def set_source_mds(self, descriptor_container: AbstractDescriptorProtocol):
        """Find the parent mds for descriptor_container and set source mds of descriptor_container."""
        mds = self.get_mds_descriptor(descriptor_container)
        descriptor_container.set_source_mds(mds.Handle)


    def disassociate_all(self,
                         entity: MultiStateEntityProtocol,
                         unbinding_mdib_version: int,
                         ignored_handle: str | None = None) -> list[str]:
        """Disassociate all associated states in entity.

        The method returns a list of states that were disassociated.
        :param entity: ProviderMultiStateEntity
        :param ignored_handle: the context state with this Handle shall not be touched.
        """
        pm_types = self._mdib.data_model.pm_types
        disassociated_state_handles = []
        for state in entity.states.values():
            if state.Handle == ignored_handle or state.ContextAssociation == pm_types.ContextAssociation.NO_ASSOCIATION:
                # If state is already part of this transaction leave it also untouched, accept what the user wanted.
                # If state is not associated, also do not touch it.
                continue
            if state.ContextAssociation != pm_types.ContextAssociation.DISASSOCIATED \
                    or state.UnbindingMdibVersion is None:
                state.ContextAssociation = pm_types.ContextAssociation.DISASSOCIATED
                if state.UnbindingMdibVersion is None:
                    state.UnbindingMdibVersion = unbinding_mdib_version
                    state.BindingEndTime = time.time()
                disassociated_state_handles.append(state.Handle)
        return disassociated_state_handles

class DescriptorFactory:
    """DescriptorFactory provides some methods to make creation of descriptors easier."""

    def __init__(self, mdib: ProviderMdib):
        self._mdib = mdib

    @staticmethod
    def _create_descriptor_container(container_cls: type,
                                     handle: str,
                                     parent_handle: str,
                                     coded_value: CodedValue,
                                     safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        obj = container_cls(handle=handle, parent_handle=parent_handle)
        obj.SafetyClassification = safety_classification
        obj.Type = coded_value
        return obj

    def create_vmd_descriptor_container(self,
                                        handle: str,
                                        parent_handle: str,
                                        coded_value: CodedValue,
                                        safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        """Create an VmdDescriptorContainer with the given properties.

        :param handle: Handle of the new container
        :param parent_handle: Handle of the parent
        :param coded_value: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value
        :return: the created object
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.VmdDescriptor)
        return self._create_descriptor_container(cls, handle, parent_handle, coded_value, safety_classification)

    def create_channel_descriptor_container(self,
                                            handle: str,
                                            parent_handle: str,
                                            coded_value: CodedValue,
                                            safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        """Create a ChannelDescriptorContainer with the given properties.

        :param handle: Handle of the new container.
        :param parent_handle: Handle of the parent.
        :param coded_value: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value.
        :return: the created object.
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.ChannelDescriptor)
        return self._create_descriptor_container(cls, handle, parent_handle, coded_value, safety_classification)

    def create_string_metric_descriptor_container(self,  # noqa: PLR0913
                                                  handle: str,
                                                  parent_handle: str,
                                                  coded_value: CodedValue,
                                                  safety_classification: SafetyClassification,
                                                  unit: CodedValue,
                                                  metric_availability: MetricAvailability,
                                                  metric_category: MetricCategory) -> AbstractDescriptorProtocol:
        """Create a StringMetricDescriptorContainer with the given properties.

        :param handle: Handle of the new container
        :param parent_handle: Handle of the parent
        :param coded_value: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value
        :param unit: a CodedValue
        :param metric_availability: pmtypes.MetricAvailability
        :param metric_category: pmtypes.MetricCategory
        :return: the created object
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.StringMetricDescriptor)
        obj = self._create_descriptor_container(cls, handle, parent_handle, coded_value, safety_classification)
        obj.Unit = unit
        obj.MetricAvailability = metric_availability
        obj.MetricCategory = metric_category
        return obj

    def create_enum_string_metric_descriptor_container(self,  # noqa: PLR0913
                                                       handle: str,
                                                       parent_handle: str,
                                                       coded_value: CodedValue,
                                                       safety_classification: SafetyClassification,
                                                       unit: CodedValue,
                                                       allowed_values: Any,
                                                       metric_availability: MetricAvailability,
                                                       metric_category: MetricCategory) -> AbstractDescriptorProtocol:
        """Create an EnumStringMetricDescriptorContainer with the given properties.

        :param handle: Handle of the new container
        :param parent_handle: Handle of the parent
        :param coded_value: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value
        :param unit: pmtypes.CodedValue
        :param allowed_values:
        :param metric_availability: pmtypes.MetricAvailability
        :param metric_category: pmtypes.MetricCategory
        :return: the created object.
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.EnumStringMetricDescriptor)
        obj = self._create_descriptor_container(cls, handle, parent_handle, coded_value, safety_classification)
        obj.Unit = unit
        obj.MetricAvailability = metric_availability
        obj.MetricCategory = metric_category
        obj.AllowedValue = allowed_values
        return obj

    def create_clock_descriptor_container(self,
                                          handle: str,
                                          parent_handle: str,
                                          coded_value: CodedValue,
                                          safety_classification: SafetyClassification) -> AbstractDescriptorProtocol:
        """Create a ClockDescriptorContainer with the given properties.

        :param handle: Handle of the new container
        :param parent_handle: Handle of the parent
        :param coded_value: a pmtypes.CodedValue instance that defines what this object represents in medical terms.
        :param safety_classification: a pmtypes.SafetyClassification value
        :return: the created object.
        """
        model = self._mdib.data_model
        cls = model.get_descriptor_container_class(model.pm_names.ClockDescriptor)
        return self._create_descriptor_container(cls, handle, parent_handle, coded_value, safety_classification)



from __future__ import annotations

from typing import TYPE_CHECKING

from sdc11073.namespaces import PrefixesEnum
from sdc11073.xml_types.addressing_types import HeaderInformationBlock

from .serviceclientbase import GetRequestResult, HostedServiceClient

if TYPE_CHECKING:
    from sdc11073.consumer.manipulator import RequestManipulatorProtocol


class GetServiceClient(HostedServiceClient):
    """Client for GetService."""

    port_type_name = PrefixesEnum.SDC.tag('GetService')

    def get_mdib(self, request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetMdib request."""
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetMdib()
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        result = received_message_data.msg_reader.read_get_mdib_response(received_message_data)
        return GetRequestResult(received_message_data, result)

    def get_md_description(self, requested_handles: list[str] | None = None,
                           request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetMdDescription request.

        :param requested_handles: None if all states shall be requested, otherwise a list of handles
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetMdDescription()
        if requested_handles is not None:
            request.HandleRef.extend(requested_handles)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetMdDescriptionResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)

        return GetRequestResult(received_message_data, report)

    def get_md_state(self, requested_handles: list[str] | None = None,
                     request_manipulator: RequestManipulatorProtocol | None = None) -> GetRequestResult:
        """Send a GetMdState request.

        :param requested_handles: None if all states shall be requested, otherwise a list of handles
        :param request_manipulator: see documentation of RequestManipulatorProtocol
        """
        data_model = self._sdc_definitions.data_model
        request = data_model.msg_types.GetMdState()
        if requested_handles is not None:
            request.HandleRef.extend(requested_handles)
        inf = HeaderInformationBlock(action=request.action, addr_to=self.endpoint_reference.Address)
        message = self._msg_factory.mk_soap_message(inf, payload=request)
        received_message_data = self.post_message(message, request_manipulator=request_manipulator)
        cls = data_model.msg_types.GetMdStateResponse
        report = cls.from_node(received_message_data.p_msg.msg_node)
        return GetRequestResult(received_message_data, report)



from .porttypebase import DPWSPortTypeBase, WSDLMessageDescription, WSDLOperationBinding, mk_wsdl_two_way_operation
from .porttypebase import msg_prefix
from sdc11073.dispatch import DispatchKey
from sdc11073.exceptions import FunctionNotImplementedError
from sdc11073.namespaces import PrefixesEnum
from sdc11073.pysoap.soapenvelope import Fault, faultcodeEnum


class ContainmentTreeService(DPWSPortTypeBase):
    port_type_name = PrefixesEnum.SDC.tag('ContainmentTreeService')
    WSDLMessageDescriptions = (WSDLMessageDescription('GetDescriptor',
                                                      (f'{msg_prefix}:GetDescriptor',)),
                               WSDLMessageDescription('GetDescriptorResponse',
                                                      (f'{msg_prefix}:GetDescriptorResponse',)),
                               WSDLMessageDescription('GetContainmentTree',
                                                      (f'{msg_prefix}:GetContainmentTreeResponse',)),
                               WSDLMessageDescription('GetContainmentTreeResponse',
                                                      (f'{msg_prefix}:GetContainmentTreeResponse',)),
                               )
    WSDLOperationBindings = (WSDLOperationBinding('GetDescriptor', 'literal', 'literal'),
                             WSDLOperationBinding('GetContainmentTree', 'literal', 'literal'))

    def register_hosting_service(self, hosting_service):
        super().register_hosting_service(hosting_service)
        actions = self._mdib.sdc_definitions.Actions
        msg_names = self._mdib.sdc_definitions.data_model.msg_names
        hosting_service.register_post_handler(DispatchKey(actions.GetContainmentTree, msg_names.GetContainmentTree),
                                              self._on_get_containment_tree)
        hosting_service.register_post_handler(DispatchKey(actions.GetDescriptor, msg_names.GetDescriptor),
                                              self._on_get_descriptor)

    def _on_get_containment_tree(self, request_data):  # pylint: disable=no-self-use
        # ToDo: implement, currently method only raises a soap fault
        fault = Fault()
        fault.Code.Value = faultcodeEnum.RECEIVER
        fault.add_reason_text('not implemented')

        raise FunctionNotImplementedError(fault)

    def _on_get_descriptor(self, request_data):  # pylint: disable=no-self-use
        # ToDo: implement, currently method only raises a soap fault
        fault = Fault()
        fault.Code.Value = faultcodeEnum.RECEIVER
        fault.add_reason_text('not implemented')
        raise FunctionNotImplementedError(fault)

    def add_wsdl_port_type(self, parent_node):
        port_type = self._mk_port_type_node(parent_node)
        mk_wsdl_two_way_operation(port_type, operation_name='GetDescriptor')
        mk_wsdl_two_way_operation(port_type, operation_name='GetContainmentTree')



# noqa: D104
from sdc11073.xml_types.wsd_types import ScopesType  # make ScopesType visible in wsdiscovery for convenience

from .wsdimpl import (
    WSDiscovery,
    WSDiscoverySingleAdapter,
)

__all__ = ['WSDiscovery', 'WSDiscoverySingleAdapter', 'ScopesType']



"""Get the hosts network adapters and ip addresses."""
from __future__ import annotations

import ipaddress

import ifaddr

IP_BLACKLIST = ('0.0.0.0',  # noqa: S104
                None)  # None can happen if an adapter does not have any IP address associated


class NetworkAdapterNotFoundError(Exception):
    """Exception when no network adapter is found."""

    def __init__(self, ip: ipaddress.IPv4Address, *args, **kwargs):  # noqa: ANN002 ANN003
        super().__init__(args, kwargs)
        self.ip = ip


class NetworkAdapter(ipaddress.IPv4Interface):
    """Represents a network adapter."""

    def __init__(self, name: str, description: str, address: str, network_prefix: str | int):
        """Create a network adapter instance.

        :param name: name of the interface
        :param description: descriptive, more general name of the network adapter
        :param address: ip address of the network adapter
        :param network_prefix: network prefix
        """
        super().__init__(f'{address}/{network_prefix}')
        self.name: str = name
        self.description: str = description

    def __str__(self) -> str:
        return f'{self.name}: {super().__str__()}'

    def __repr__(self) -> str:
        return f'{self.name}: {super().__str__()} ({self.description})'


def get_adapters() -> list[NetworkAdapter]:
    """Get all active and connected host network adapters.

    :return: list of enabled and connected network adapters on the host
    """
    adapters: list[NetworkAdapter] = []
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            if ip.is_IPv4 and ip.ip not in IP_BLACKLIST:
                adapters.append(NetworkAdapter(name=ip.nice_name,
                                               description=adapter.nice_name,
                                               address=ip.ip,
                                               network_prefix=ip.network_prefix))

    return adapters


def get_adapter_containing_ip(ip: ipaddress.IPv4Address | str) -> NetworkAdapter:
    """Get host network adapter containing the specified ip address in its network range.

    :param ip: ip address from which the adapter is to be determined
    :return: host network adapter containing the specified ip address in its network range
    :raise NetworkAdapterNotFoundError: no network adapter contains the specified ip
    :raise RuntimeError: multiple network adapter containing the specified ip
    """
    ip = ipaddress.IPv4Address(ip) if isinstance(ip, str) else ip

    adapters = get_adapters()
    filtered_adapters = [adapter for adapter in adapters if ip in adapter.network]

    if not filtered_adapters:
        raise NetworkAdapterNotFoundError(ip, f'No network adapter contains ip address "{ip}". Detected {adapters}')

    if len(filtered_adapters) > 1:
        # any ip address could be taken but do not choose randomly
        # sort ip addresses to determine the closest one
        filtered_adapters.sort(key=lambda a: abs(int(a) - int(ip)))

    return filtered_adapters[0]




def apply_map(function, *iterable):
    """call function for all elements of iterable(s).
    apply_map uses builtin map internally."""
    return list(map(function, *iterable))


def _short_action_string(action: str):
    """ return only the last 2 elements of the action """
    elements = action.split('/')
    ret = '/'.join(elements[-2:])
    return ret


def short_filter_string(actions):
    """
    Helper function to make shorter action strings for logging
    :param actions: list of strings
    :return: a comma separated string of shortened names
    """
    return ', '.join([_short_action_string(a) for a in actions])



from __future__ import annotations

from collections import defaultdict, namedtuple
from threading import RLock
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from collections.abc import Iterable

"""
This module implements an in-memory table with indices for faster access to objects.
Example: You have a class
class Person:
    def __init__(self, first_name, last_name, age):
        self.first_name = first_name
        self.last_name = last_name
        self._age = age

    def getAge(self):
        return int(self._age

you can setup an in-memory table like this:
person_lookup = MultiKeyLookup()
person_lookup.add_index('by_firstname', multikey.IndexDefinition(lambda obj: obj.first_name))
person_lookup.add_index('by_firstname', multikey.IndexDefinition(lambda obj: obj.last_name))
person_lookup.add_index('by_age', multikey.IndexDefinition(lambda obj: obj.getAge()))

person_lookup.add_object(Person('Peter', 'Miller', 42)
person_lookup.add_object(Person('John', 'Myers', 50)
person_lookup.add_object(Person('Agnes', 'Miller', 42)

accessing by index:
all_millers = person_lookup.by_lastname.get('Miller')
all_42_agers = person_lookup.by_age.get(42)
"""



class IndexDefinition(dict):
    """An index allows to group objects by values.

    This is a dictionary that has lists ob objects as value.
    Each list contains objects that have the same key member.
    """

    def __init__(self, get_key_func:Callable[[Any], Any], index_none_values: bool=True):
        """Construct an index.

        :param get_key_func: a callable that returns a key value from a given object
        :param index_none_values: if True, a None key is handled like every other value.
                                if False,a None key is not added to index.
        """
        super().__init__()
        self._get_key_func = get_key_func
        self._index_none_values = index_none_values
        self._lock: RLock | None = None

    def get_one(self, key: Any, allow_none: bool=False) -> Any | None:
        """Return exactly one object instead of a list (like get method does).

        It raises a ValueError if there are multiple values available for the key.
        It raises a KeyError if allow_none is False and the key is not present.
        :param key:
        :param allow_none:
        :return:
        """
        with self._lock:
            if allow_none:
                result = self.get(key)
                if result is None:
                    return result
            else:
                result = self[key]
            if len(result) > 1:
                raise ValueError(f'get_one: key "{key}" has {len(result)} objects')
            return result[0]

    def get(self, *args, **kwargs) -> list[Any] | None:
        """Overwritten get method that uses lock."""
        with self._lock:
            return super().get(*args, **kwargs)

    def __getitem__(self, key) -> list[Any] | None:
        """Overwritten __getitem__ method that uses lock."""
        with self._lock:
            return super().__getitem__(key)

    def set_lock(self, lock: RLock):
        """Set the lock to be used."""
        self._lock = lock

    def mk_keys(self, obj: Any) -> list[Any] | None:
        """Determine key for obj and add it to list in self[key]."""
        key = self._get_key_func(obj)
        if not self._index_none_values and key is None:
            return None
        try:
            self[key].append(obj)
        except KeyError:
            self[key] = [obj]
        return [key]

    def rm_key(self, key: Any, obj: Any):
        """Remove obj from list self[key]."""
        try:
            obj_list = self[key]
            obj_list.remove(obj)
            if len(obj_list) == 0:
                del self[key]
        except (KeyError, ValueError):
            pass


class UIndexDefinition(IndexDefinition):
    """A unique Index, there can only be one object with that key."""

    def mk_keys(self, obj: Any) -> list[Any] | None:
        """Determine key for obj and add it to list in self[key].

        If key is already known, raise a KeyError.
        """
        keys = self._get_key_func(obj)
        if not self._index_none_values and keys is None:
            return None
        if isinstance(keys, list):
            raise ValueError(f'list of keys not allowed in UIndex: obj={obj}, keys={keys}')
        keys = [keys]
        for k in keys:
            if k in self:
                raise KeyError(f'key "{k}" in already in this UIndex')
            self[k] = [obj]
        return keys


class IndexDefinition1n(IndexDefinition):
    """Index for member values that are a list of keys (1:n relationship)."""

    def mk_keys(self, obj: Any) -> list[Any] | None:
        """Determine key for obj and add it to list in self[key]."""
        keys = self._get_key_func(obj)
        if not self._index_none_values and keys is None:
            return None
        for k in keys:
            try:
                self[k].append(obj)
            except KeyError:
                self[k] = [obj]
        return keys


class ObjectSelector:
    """Implements a mechanism to filter objects."""

    def __init__(self, selected_objects: Iterable[Any]):
        self.objects = selected_objects

    def find(self, **kwargs) -> ObjectSelector:
        """Return an ObjectSelector with a subset of data.

        The filter in kwargs implements an OR combination of args.
        Values are compared for equality (==), not identity (is).

        Example:
        -------
        - table contains persons with first_name and family_name.
        - find(first_name='Mike') returns all persons with first_name == 'Mike'.
        """
        result = []
        for obj in self.objects:
            for name, value in kwargs.items():
                try:
                    val = getattr(obj, name)
                    if callable(val):
                        val = val()
                except AttributeError:
                    pass
                else:
                    if val == value:
                        result.append(obj)
                        break
        return ObjectSelector(result)


_ObjRef = namedtuple('_ObjRef', 'index_dict key')  # used internally in MultiKeyLookup to keep track of all indexes.


# when we remove an object we need it to delete all indices referencing it

class MultiKeyLookup:
    """A combination of a list of objects and dictionaries.

    This mimics a database table with multiple indices.
    The dictionaries can be used to directly access values by a key.
    """

    def __init__(self):
        self._objects = set()  # contains the objects
        self._object_ids = defaultdict(
            list)  # key = id, value = list of ((_idx_defs, key) tuples that reference the object
        self._idx_defs = {}  # holds UIndexDefinition Objects
        self._lock = RLock()

    @property
    def objects(self) -> set[Any]:
        """Return a set of all objects in table."""
        return self._objects

    @property
    def lock(self) -> RLock:
        """Return the lock of the table."""
        return self._lock

    def __getattr__(self, name: str):
        return self._idx_defs[name]

    def add_index(self, index_name: str, index_definition: IndexDefinition):
        """Add index to table."""
        self._idx_defs[index_name] = index_definition
        index_definition.set_lock(self._lock)
        # add existing objects to new lookup
        for obj in self._objects:
            keys = index_definition.mk_keys(obj)
            for k in keys:
                self._object_ids[id(obj)].append(_ObjRef(index_definition, k))

    def add_object(self, obj: Any):
        """Add object to table.

        Indices are updated accordingly.
        """
        if obj in self._objects:
            return
        with self._lock:
            self._objects.add(obj)
            self._mk_indices(obj)

    def add_object_no_lock(self, obj: Any):
        """Add object to table without using the lock.

        Indices are updated accordingly.
        """
        if obj in self._objects:
            return
        self._objects.add(obj)
        self._mk_indices(obj)

    def add_objects(self, objects: list[Any]):
        """Add objects to table.

        Indices are updated accordingly.
        """
        with self._lock:
            self.add_objects_no_lock(objects)

    def add_objects_no_lock(self, objects: list[Any]):
        """Add objects to table without using the lock.

        Indices are updated accordingly.
        """
        for obj in objects:
            if obj in self._objects:
                continue
            self._objects.add(obj)
            self._mk_indices(obj)

    def _mk_indices(self, obj: Any):
        all_keys = []  # for this object
        for index_definition in self._idx_defs.values():
            try:
                keys = index_definition.mk_keys(obj)
                for k in keys:
                    all_keys.append(_ObjRef(index_definition, k))
            except (TypeError, AttributeError):
                pass
        self._object_ids[id(obj)].extend(all_keys)

    def _rm_indices(self, obj: Any):
        obj_refs = self._object_ids.get(id(obj), [])
        for obj_ref in obj_refs:
            obj_ref.index_dict.rm_key(obj_ref.key, obj)
        del self._object_ids[id(obj)]

    def remove_object(self, obj: Any):
        """Remove object from table.

        Indices are updated accordingly.
        """
        obj_refs = self._object_ids.get(id(obj))
        if obj_refs is None:
            return
        with self._lock:
            self._rm_indices(obj)
            self._objects.remove(obj)

    def remove_object_no_lock(self, obj: Any):
        """Remove object from table without using lock.

        Indices are updated accordingly.
        """
        obj_refs = self._object_ids.get(id(obj))
        if obj_refs is None:
            return
        self._rm_indices(obj)
        self._objects.remove(obj)

    def remove_objects(self, objects: list[Any]):
        """Remove objects from table.

        Indices are updated accordingly.
        """
        with self._lock:
            self.remove_objects_no_lock(objects)

    def remove_objects_no_lock(self, objects: list[Any]):
        """Remove objects from table without using lock.

        Indices are updated accordingly.
        """
        for obj in objects:
            obj_refs = self._object_ids.get(id(obj))
            if obj_refs is None:
                continue
            self._rm_indices(obj)
            self._objects.remove(obj)

    def update_object(self, obj: Any):
        """Update indices according to current values in obj."""
        if obj not in self._objects:
            raise ValueError(f'object {obj} not known')
        with self._lock:
            self._rm_indices(obj)
            self._mk_indices(obj)

    def update_object_no_lock(self, obj: Any):
        """Update indices according to current values in obj without using lock."""
        if obj not in self._objects:
            raise ValueError(f'object {obj} not known')
        self._rm_indices(obj)
        self._mk_indices(obj)

    def update_objects(self, objs: list[Any]):
        """Update indices according to current values in objs."""
        with self._lock:
            self.update_objects_no_lock(objs)

    def update_objects_no_lock(self, objs: list[Any]):
        """Update indices according to current values in objs without using lock."""
        for obj in objs:
            if obj not in self._objects:
                raise ValueError(f'object {obj} not known')
            with self._lock:
                self._rm_indices(obj)
                self._mk_indices(obj)

    def clear(self):
        """Remove all objects from table."""
        with self._lock:
            for index_definition in self._idx_defs.values():
                index_definition.clear()
            self._object_ids.clear()
            self._objects.clear()

    def find(self, **kwargs) -> ObjectSelector:
        """Return an ObjectSelector with a subset of all objects that match the filter criteria in kwargs."""
        sel = ObjectSelector(self._objects)
        with self._lock:
            return sel.find(**kwargs)

    def find_no_lock(self, **kwargs) -> ObjectSelector:
        """Like find, but without using lock."""
        sel = ObjectSelector(self._objects)
        return sel.find(**kwargs)
